# -*- coding: utf-8 -*-
"""
    pygments.styles.inkpot
    ~~~~~~~~~~~~~~~~~~~~~~

    A highlighting style for Pygments, inspired by the Inkpot theme for VIM.

    :copyright: Copyright 2006-2019 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.style import Style
from pygments.token import Text, Other, Keyword, Name, Comment, String, \
    Error, Number, Operator, Generic, Whitespace, Punctuation


class InkPotStyle(Style):
    background_color = "#1e1e27"
    default_style = ""
    styles = {
        Text:                      "#cfbfad",
        Other:                     "#cfbfad",
        Whitespace:                "#434357",
        Comment:                   "#cd8b00",
        Comment.Preproc:           "#409090",
        Comment.PreprocFile:       "bg:#404040 #ffcd8b",
        Comment.Special:           "#808bed",

        Keyword:                   "#808bed",
        Keyword.Pseudo:            "nobold",
        Keyword.Type:              "#ff8bff",

        Operator:                  "#666666",

        Punctuation:               "#cfbfad",

        Name:                      "#cfbfad",
        Name.Attribute:            "#cfbfad",
        Name.Builtin.Pseudo:       '#ffff00',
        Name.Builtin:              "#808bed",
        Name.Class:                "#ff8bff",
        Name.Constant:             "#409090",
        Name.Decorator:            "#409090",
        Name.Exception:            "#ff0000",
        Name.Function:             "#c080d0",
        Name.Label:                "#808bed",
        Name.Namespace:            "#ff0000",
        Name.Variable:             "#cfbfad",

        String:                    "bg:#404040 #ffcd8b",
        String.Doc:                "#808bed",

        Number:                    "#f0ad6d",

        Generic.Heading:           "bold #000080",
        Generic.Subheading:        "bold #800080",
        Generic.Deleted:           "#A00000",
        Generic.Inserted:          "#00A000",
        Generic.Error:             "#FF0000",
        Generic.Emph:              "italic",
        Generic.Strong:            "bold",
        Generic.Prompt:            "bold #000080",
        Generic.Output:            "#888",
        Generic.Traceback:         "#04D",

        Error:                     "bg:#6e2e2e #ffffff"
    }



# -*- coding: utf-8 -*-
"""
requests_toolbelt.adapters.host_header_ssl
==========================================

This file contains an implementation of the HostHeaderSSLAdapter.
"""

from requests.adapters import HTTPAdapter


class HostHeaderSSLAdapter(HTTPAdapter):
    """
    A HTTPS Adapter for Python Requests that sets the hostname for certificate
    verification based on the Host header.

    This allows requesting the IP address directly via HTTPS without getting
    a "hostname doesn't match" exception.

    Example usage:

        >>> s.mount('https://', HostHeaderSSLAdapter())
        >>> s.get("https://93.184.216.34", headers={"Host": "example.org"})

    """

    def send(self, request, **kwargs):
        # HTTP headers are case-insensitive (RFC 7230)
        host_header = None
        for header in request.headers:
            if header.lower() == "host":
                host_header = request.headers[header]
                break

        connection_pool_kwargs = self.poolmanager.connection_pool_kw

        if host_header:
            connection_pool_kwargs["assert_hostname"] = host_header
        elif "assert_hostname" in connection_pool_kwargs:
            # an assert_hostname from a previous request may have been left
            connection_pool_kwargs.pop("assert_hostname", None)

        return super(HostHeaderSSLAdapter, self).send(request, **kwargs)



from __future__ import absolute_import, division, unicode_literals

import socket

from .base import StatsClientBase, PipelineBase


class Pipeline(PipelineBase):

    def __init__(self, client):
        super(Pipeline, self).__init__(client)
        self._maxudpsize = client._maxudpsize

    def _send(self):
        data = self._stats.popleft()
        while self._stats:
            # Use popleft to preserve the order of the stats.
            stat = self._stats.popleft()
            if len(stat) + len(data) + 1 >= self._maxudpsize:
                self._client._after(data)
                data = stat
            else:
                data += '\n' + stat
        self._client._after(data)


class StatsClient(StatsClientBase):
    """A client for statsd."""

    def __init__(self, host='localhost', port=8125, prefix=None,
                 maxudpsize=512, ipv6=False):
        """Create a new client."""
        fam = socket.AF_INET6 if ipv6 else socket.AF_INET
        family, _, _, _, addr = socket.getaddrinfo(
            host, port, fam, socket.SOCK_DGRAM)[0]
        self._addr = addr
        self._sock = socket.socket(family, socket.SOCK_DGRAM)
        self._prefix = prefix
        self._maxudpsize = maxudpsize

    def _send(self, data):
        """Send data to statsd."""
        try:
            self._sock.sendto(data.encode('ascii'), self._addr)
        except (socket.error, RuntimeError):
            # No time for love, Dr. Jones!
            pass

    def pipeline(self):
        return Pipeline(self)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Helpers to test your application's logging behavior.

.. versionadded:: 20.1.0

See :doc:`testing`.
"""

from contextlib import contextmanager

from ._config import configure, get_config
from .exceptions import DropEvent


__all__ = ["LogCapture", "capture_logs"]


class LogCapture(object):
    """
    Class for capturing log messages in its entries list.
    Generally you should use `structlog.testing.capture_logs`,
    but you can use this class if you want to capture logs with other patterns.

    .. versionadded:: 20.1.0
    """

    def __init__(self):
        self.entries = []

    def __call__(self, _, method_name, event_dict):
        event_dict["log_level"] = method_name
        self.entries.append(event_dict)
        raise DropEvent


@contextmanager
def capture_logs():
    """
    Context manager that appends all logging statements to its yielded list
    while it is active.

    Attention: this is **not** thread-safe!

    .. versionadded:: 20.1.0
    """
    cap = LogCapture()
    old_processors = get_config()["processors"]
    try:
        configure(processors=[cap])
        yield cap.entries
    finally:
        configure(processors=old_processors)


class ReturnLoggerFactory(object):
    r"""
    Produce and cache `ReturnLogger`\ s.

    To be used with `structlog.configure`\ 's *logger_factory*.

    Positional arguments are silently ignored.

    .. versionadded:: 0.4.0
    """

    def __init__(self):
        self._logger = ReturnLogger()

    def __call__(self, *args):
        return self._logger


class ReturnLogger(object):
    """
    Return the arguments that it's called with.

    >>> from structlog import ReturnLogger
    >>> ReturnLogger().msg("hello")
    'hello'
    >>> ReturnLogger().msg("hello", when="again")
    (('hello',), {'when': 'again'})

    .. versionchanged:: 0.3.0
        Allow for arbitrary arguments and keyword arguments to be passed in.
    """

    def msg(self, *args, **kw):
        """
        Return tuple of ``args, kw`` or just ``args[0]`` if only one arg passed
        """
        # Slightly convoluted for backwards compatibility.
        if len(args) == 1 and not kw:
            return args[0]
        else:
            return args, kw

    log = debug = info = warn = warning = msg
    fatal = failure = err = error = critical = exception = msg



# -*- coding: utf-8 -*-
# $Id: ja.py 4564 2006-05-21 20:44:42Z wiemann $
# Author: Hisashi Morita <hisashim@kt.rim.or.jp>
# Copyright: This module has been placed in the public domain.

# New language mappings are welcome.  Before doing a new translation, please
# read <http://docutils.sf.net/docs/howto/i18n.html>.  Two files must be
# translated for each language: one in docutils/languages, the other in
# docutils/parsers/rst/languages.

"""
Japanese-language mappings for language-dependent features of Docutils.
"""

__docformat__ = 'reStructuredText'

labels = {
      # fixed: language-dependent
      'author': u'著者',
      'authors': u'著者',
      'organization': u'組織',
      'address': u'住所',
      'contact': u'連絡先',
      'version': u'バージョン',
      'revision': u'リビジョン',
      'status': u'ステータス',
      'date': u'日付',
      'copyright': u'著作権',
      'dedication': u'献辞',
      'abstract': u'概要',
      'attention': u'注目!',
      'caution': u'注意!',
      'danger': u'!危険!',
      'error': u'エラー',
      'hint': u'ヒント',
      'important': u'重要',
      'note': u'備考',
      'tip': u'通報',
      'warning': u'警告',
      'contents': u'目次'}
"""Mapping of node class name to label text."""

bibliographic_fields = {
      # language-dependent: fixed
      u'著者': 'author',
      u' n/a': 'authors',
      u'組織': 'organization',
      u'住所': 'address',
      u'連絡先': 'contact',
      u'バージョン': 'version',
      u'リビジョン': 'revision',
      u'ステータス': 'status',
      u'日付': 'date',
      u'著作権': 'copyright',
      u'献辞': 'dedication',
      u'概要': 'abstract'}
"""Japanese (lowcased) to canonical name mapping for bibliographic fields."""

author_separators = [';', ',']
"""List of separator strings for the 'Authors' bibliographic field. Tried in
order."""



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Processors useful regardless of the logging framework.
"""

from __future__ import absolute_import, division, print_function

import datetime
import json
import operator
import sys
import time

import six

from structlog._frames import (
    _find_first_app_frame_and_name,
    _format_exception,
    _format_stack,
)


class KeyValueRenderer(object):
    """
    Render ``event_dict`` as a list of ``Key=repr(Value)`` pairs.

    :param bool sort_keys: Whether to sort keys when formatting.
    :param list key_order: List of keys that should be rendered in this exact
        order.  Missing keys will be rendered as ``None``, extra keys depending
        on *sort_keys* and the dict class.
    :param bool drop_missing: When ``True``, extra keys in *key_order* will be
        dropped rather than rendered as ``None``.
    :param bool repr_native_str: When ``True``, :func:`repr()` is also applied
        to native strings (i.e. unicode on Python 3 and bytes on Python 2).
        Setting this to ``False`` is useful if you want to have human-readable
        non-ASCII output on Python 2.

    .. versionadded:: 0.2.0 *key_order*
    .. versionadded:: 16.1.0 *drop_missing*
    .. versionadded:: 17.1.0 *repr_native_str*
    """

    def __init__(
        self,
        sort_keys=False,
        key_order=None,
        drop_missing=False,
        repr_native_str=True,
    ):
        # Use an optimized version for each case.
        if key_order and sort_keys:

            def ordered_items(event_dict):
                items = []
                for key in key_order:
                    value = event_dict.pop(key, None)
                    if value is not None or not drop_missing:
                        items.append((key, value))
                items += sorted(event_dict.items())
                return items

        elif key_order:

            def ordered_items(event_dict):
                items = []
                for key in key_order:
                    value = event_dict.pop(key, None)
                    if value is not None or not drop_missing:
                        items.append((key, value))
                items += event_dict.items()
                return items

        elif sort_keys:

            def ordered_items(event_dict):
                return sorted(event_dict.items())

        else:
            ordered_items = operator.methodcaller("items")

        self._ordered_items = ordered_items

        if repr_native_str is True:
            self._repr = repr
        else:

            def _repr(inst):
                if isinstance(inst, str):
                    return inst
                else:
                    return repr(inst)

            self._repr = _repr

    def __call__(self, _, __, event_dict):
        return " ".join(
            k + "=" + self._repr(v) for k, v in self._ordered_items(event_dict)
        )


class UnicodeEncoder(object):
    """
    Encode unicode values in ``event_dict``.

    :param str encoding: Encoding to encode to (default: ``"utf-8"``).
    :param str errors: How to cope with encoding errors (default
        ``"backslashreplace"``).

    Useful if you're running Python 2 as otherwise ``u"abc"`` will be rendered
    as ``'u"abc"'``.

    Just put it in the processor chain before the renderer.
    """

    def __init__(self, encoding="utf-8", errors="backslashreplace"):
        self._encoding = encoding
        self._errors = errors

    def __call__(self, logger, name, event_dict):
        for key, value in event_dict.items():
            if isinstance(value, six.text_type):
                event_dict[key] = value.encode(self._encoding, self._errors)
        return event_dict


class UnicodeDecoder(object):
    """
    Decode byte string values in ``event_dict``.

    :param str encoding: Encoding to decode from (default: ``"utf-8"``).
    :param str errors: How to cope with encoding errors (default:
        ``"replace"``).

    Useful if you're running Python 3 as otherwise ``b"abc"`` will be rendered
    as ``'b"abc"'``.

    Just put it in the processor chain before the renderer.

    .. versionadded:: 15.4.0
    """

    def __init__(self, encoding="utf-8", errors="replace"):
        self._encoding = encoding
        self._errors = errors

    def __call__(self, logger, name, event_dict):
        for key, value in event_dict.items():
            if isinstance(value, bytes):
                event_dict[key] = value.decode(self._encoding, self._errors)
        return event_dict


class JSONRenderer(object):
    """
    Render the ``event_dict`` using ``serializer(event_dict, **json_kw)``.

    :param dict json_kw: Are passed unmodified to *serializer*.  If *default*
        is passed, it will disable support for ``__structlog__``-based
        serialization.
    :param callable serializer: A :func:`json.dumps`-compatible callable that
        will be used to format the string.  This can be used to use alternative
        JSON encoders like `simplejson
        <https://pypi.org/project/simplejson/>`_ or `RapidJSON
        <https://pypi.org/project/python-rapidjson/>`_ (faster but Python
        3-only) (default: :func:`json.dumps`).

    .. versionadded:: 0.2.0
        Support for ``__structlog__`` serialization method.

    .. versionadded:: 15.4.0
        *serializer* parameter.

    .. versionadded:: 18.2.0
       Serializer's *default* parameter can be overwritten now.

    """

    def __init__(self, serializer=json.dumps, **dumps_kw):
        dumps_kw.setdefault("default", _json_fallback_handler)
        self._dumps_kw = dumps_kw
        self._dumps = serializer

    def __call__(self, logger, name, event_dict):
        return self._dumps(event_dict, **self._dumps_kw)


def _json_fallback_handler(obj):
    """
    Serialize custom datatypes and pass the rest to __structlog__ & repr().
    """
    # circular imports :(
    from structlog.threadlocal import _ThreadLocalDictWrapper

    if isinstance(obj, _ThreadLocalDictWrapper):
        return obj._dict
    else:
        try:
            return obj.__structlog__()
        except AttributeError:
            return repr(obj)


def format_exc_info(logger, name, event_dict):
    """
    Replace an ``exc_info`` field by an ``exception`` string field:

    If *event_dict* contains the key ``exc_info``, there are two possible
    behaviors:

    - If the value is a tuple, render it into the key ``exception``.
    - If the value is an Exception *and* you're running Python 3, render it
      into the key ``exception``.
    - If the value true but no tuple, obtain exc_info ourselves and render
      that.

    If there is no ``exc_info`` key, the *event_dict* is not touched.
    This behavior is analogue to the one of the stdlib's logging.
    """
    exc_info = event_dict.pop("exc_info", None)
    if exc_info:
        event_dict["exception"] = _format_exception(
            _figure_out_exc_info(exc_info)
        )
    return event_dict


class TimeStamper(object):
    """
    Add a timestamp to ``event_dict``.

    .. note::

        You should let OS tools take care of timestamping.  See also
        `logging-best-practices`.

    :param str fmt: strftime format string, or ``"iso"`` for `ISO 8601
        <https://en.wikipedia.org/wiki/ISO_8601>`_, or `None` for a `UNIX
        timestamp <https://en.wikipedia.org/wiki/Unix_time>`_.
    :param bool utc: Whether timestamp should be in UTC or local time.
    :param str key: Target key in *event_dict* for added timestamps.

    .. versionchanged:: 19.2 Can be pickled now.
    """

    __slots__ = ("_stamper", "fmt", "utc", "key")

    def __init__(self, fmt=None, utc=True, key="timestamp"):
        self.fmt, self.utc, self.key = fmt, utc, key

        self._stamper = _make_stamper(fmt, utc, key)

    def __call__(self, _, __, event_dict):
        return self._stamper(event_dict)

    def __getstate__(self):
        return {"fmt": self.fmt, "utc": self.utc, "key": self.key}

    def __setstate__(self, state):
        self.fmt = state["fmt"]
        self.utc = state["utc"]
        self.key = state["key"]

        self._stamper = _make_stamper(**state)


def _make_stamper(fmt, utc, key):
    """
    Create a stamper function.
    """
    if fmt is None and not utc:
        raise ValueError("UNIX timestamps are always UTC.")

    now = getattr(datetime.datetime, "utcnow" if utc else "now")

    if fmt is None:

        def stamper_unix(event_dict):
            event_dict[key] = time.time()
            return event_dict

        return stamper_unix
    elif fmt.upper() == "ISO":

        def stamper_iso_local(event_dict):
            event_dict[key] = now().isoformat()
            return event_dict

        def stamper_iso_utc(event_dict):
            event_dict[key] = now().isoformat() + "Z"
            return event_dict

        if utc:
            return stamper_iso_utc
        else:
            return stamper_iso_local

    def stamper_fmt(event_dict):
        event_dict[key] = now().strftime(fmt)
        return event_dict

    return stamper_fmt


def _figure_out_exc_info(v):
    """
    Depending on the Python version will try to do the smartest thing possible
    to transform *v* into an ``exc_info`` tuple.

    :rtype: tuple
    """
    if six.PY3 and isinstance(v, BaseException):
        return (v.__class__, v, v.__traceback__)
    elif isinstance(v, tuple):
        return v
    elif v:
        return sys.exc_info()

    return v


class ExceptionPrettyPrinter(object):
    """
    Pretty print exceptions and remove them from the ``event_dict``.

    :param file: Target file for output (default: ``sys.stdout``).
    :type file: file object

    This processor is mostly for development and testing so you can read
    exceptions properly formatted.

    It behaves like format_exc_info` except it removes the exception
    data from the event dictionary after printing it.

    It's tolerant to having `format_exc_info` in front of itself in the
    processor chain but doesn't require it.  In other words, it handles both
    ``exception`` as well as ``exc_info`` keys.

    .. versionadded:: 0.4.0

    .. versionchanged:: 16.0.0
       Added support for passing exceptions as ``exc_info`` on Python 3.
    """

    def __init__(self, file=None):
        if file is not None:
            self._file = file
        else:
            self._file = sys.stdout

    def __call__(self, logger, name, event_dict):
        exc = event_dict.pop("exception", None)
        if exc is None:
            exc_info = _figure_out_exc_info(event_dict.pop("exc_info", None))
            if exc_info:
                exc = _format_exception(exc_info)
        if exc:
            print(exc, file=self._file)
        return event_dict


class StackInfoRenderer(object):
    """
    Add stack information with key ``stack`` if ``stack_info`` is `True`.

    Useful when you want to attach a stack dump to a log entry without
    involving an exception.

    It works analogously to the *stack_info* argument of the Python 3 standard
    library logging but works on both 2 and 3.

    .. versionadded:: 0.4.0
    """

    def __call__(self, logger, name, event_dict):
        if event_dict.pop("stack_info", None):
            event_dict["stack"] = _format_stack(
                _find_first_app_frame_and_name()[0]
            )
        return event_dict



from __future__ import unicode_literals

import re


class URLValidator(object):
    ul = "\u00a1-\uffff"  # unicode letters range (must be a unicode string, not a raw string)

    # IP patterns
    ipv4_re = (
        r"(?:25[0-5]|2[0-4]\d|[0-1]?\d?\d)(?:\.(?:25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}"
    )
    ipv6_re = r"\[[0-9a-f:\.]+\]"  # (simple regex, validated later)

    # Host patterns
    hostname_re = (
        r"[a-z" + ul + r"0-9](?:[a-z" + ul + r"0-9-]{0,61}[a-z" + ul + r"0-9])?"
    )
    # Max length for domain name labels is 63 characters per RFC 1034 sec. 3.1
    domain_re = r"(?:\.(?!-)[a-z" + ul + r"0-9-]{1,63}(?<!-))*"
    tld_re = (
        r"\."  # dot
        r"(?!-)"  # can't start with a dash
        r"(?:[a-z" + ul + "-]{2,63}"  # domain label
        r"|xn--[a-z0-9]{1,59})"  # or punycode label
        r"(?<!-)"  # can't end with a dash
        r"\.?"  # may have a trailing dot
    )
    host_re = "(" + hostname_re + domain_re + tld_re + "|localhost)"

    netloc_re = r"(?:" + ipv4_re + "|" + ipv6_re + "|" + host_re + ")"
    port_re = r"(?::\d{2,5})?"  # port

    regex = re.compile(
        r"^(?:[a-z0-9\.\-\+]*)://"  # scheme is validated separately
        r"(?:\S+(?::\S*)?@)?"  # user:pass authentication
        + netloc_re
        + port_re
        + r"(?:[/?#][^\s]*)?"  # resource path
        r"\Z",
        re.IGNORECASE,
    )



# -*- coding: utf-8 -*-
"""
    pygments.lexers.compiled
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Just export lexer classes previously contained in this module.

    :copyright: Copyright 2006-2019 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexers.jvm import JavaLexer, ScalaLexer
from pygments.lexers.c_cpp import CLexer, CppLexer
from pygments.lexers.d import DLexer
from pygments.lexers.objective import ObjectiveCLexer, \
    ObjectiveCppLexer, LogosLexer
from pygments.lexers.go import GoLexer
from pygments.lexers.rust import RustLexer
from pygments.lexers.c_like import ECLexer, ValaLexer, CudaLexer
from pygments.lexers.pascal import DelphiLexer, Modula2Lexer, AdaLexer
from pygments.lexers.business import CobolLexer, CobolFreeformatLexer
from pygments.lexers.fortran import FortranLexer
from pygments.lexers.prolog import PrologLexer
from pygments.lexers.python import CythonLexer
from pygments.lexers.graphics import GLShaderLexer
from pygments.lexers.ml import OcamlLexer
from pygments.lexers.basic import BlitzBasicLexer, BlitzMaxLexer, MonkeyLexer
from pygments.lexers.dylan import DylanLexer, DylanLidLexer, DylanConsoleLexer
from pygments.lexers.ooc import OocLexer
from pygments.lexers.felix import FelixLexer
from pygments.lexers.nimrod import NimrodLexer
from pygments.lexers.crystal import CrystalLexer

__all__ = []



# -*- coding: utf-8 -*-
"""
    pygments.lexers.rnc
    ~~~~~~~~~~~~~~~~~~~

    Lexer for Relax-NG Compact syntax

    :copyright: Copyright 2006-2019 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexer import RegexLexer
from pygments.token import Text, Comment, Operator, Keyword, Name, String, \
    Punctuation

__all__ = ['RNCCompactLexer']


class RNCCompactLexer(RegexLexer):
    """
    For `RelaxNG-compact <http://relaxng.org>`_ syntax.

    .. versionadded:: 2.2
    """

    name = 'Relax-NG Compact'
    aliases = ['rnc', 'rng-compact']
    filenames = ['*.rnc']

    tokens = {
        'root': [
            (r'namespace\b', Keyword.Namespace),
            (r'(?:default|datatypes)\b', Keyword.Declaration),
            (r'##.*$', Comment.Preproc),
            (r'#.*$', Comment.Single),
            (r'"[^"]*"', String.Double),
            # TODO single quoted strings and escape sequences outside of
            # double-quoted strings
            (r'(?:element|attribute|mixed)\b', Keyword.Declaration, 'variable'),
            (r'(text\b|xsd:[^ ]+)', Keyword.Type, 'maybe_xsdattributes'),
            (r'[,?&*=|~]|>>', Operator),
            (r'[(){}]', Punctuation),
            (r'.', Text),
        ],

        # a variable has been declared using `element` or `attribute`
        'variable': [
            (r'[^{]+', Name.Variable),
            (r'\{', Punctuation, '#pop'),
        ],

        # after an xsd:<datatype> declaration there may be attributes
        'maybe_xsdattributes': [
            (r'\{', Punctuation, 'xsdattributes'),
            (r'\}', Punctuation, '#pop'),
            (r'.', Text),
        ],

        # attributes take the form { key1 = value1 key2 = value2 ... }
        'xsdattributes': [
            (r'[^ =}]', Name.Attribute),
            (r'=', Operator),
            (r'"[^"]*"', String.Double),
            (r'\}', Punctuation, '#pop'),
            (r'.', Text),
        ],
    }



import logging
from power_switch import pybreaker
from structlog import get_logger
import exceptions

_breakers_map = {}


class Failsafe(object):
    """
    Utility class for creating circuit breakers.

    It is important to notice that every function decorated
    via failsafe must have a param named `key`. The value of
    this param is used to uniquely identify the circuit breaker
    instance. The callee must provide the value of this `key`.


    Basic usage::

        # Create a global failsafe instance
        failsafe = Failsafe(service_name='YourServiceName')

        # Use breaker decorator on external/dangerous calls
        @failsafe.breaker()
        def external_call(key=None):
            # perform dangerous operations


    To use ``redis`` for storing the circuit breaker state::

        failsafe = Failsafe(
            service_name='YourServiceName'
            redis_client=redis_client_object
        )


    A more advanced use case of breaker decorator::

        @failsafe.breaker(err_threshold=0.5, request_volume_window=10, reset_timeout=60)
        def external_call(key=None):
            # perform dangerous operations

    `err_threshold`: err_rate after which circuit open, this
                    value lies between 0 to 1
    `reset_timeout`: time in seconds before the circuit breaker
                   state changes to half open
    `request_volume_window` : box to determine, how many request are
                   taken into consideration for calculation
                   of current err_rate
    """

    def __init__(
        self,
        service_name,
        logger=None
    ):
        """
        :param service_name: str, name of the service
        :param redis_client: redis.Redis, if provided this circuit breaker uses
                redis for storing the state

        """
        self._service_name = service_name
        if logger:
            self.logger = logger
        else:
            log = logging.getLogger(__name__)
            console_handler = logging.StreamHandler()
            console_handler.setLevel(logging.INFO)
            console_handler_formatter = logging.Formatter(
                '[%(asctime)s] [%(levelname)-8s] [%(name)s:%(lineno)s] -- %('
                'message)s'
            )
            console_handler.setFormatter(console_handler_formatter)
            log.addHandler(console_handler)
            log.setLevel(logging.INFO)
            self.logger = get_logger().new()

    def breaker(self, err_threshold=0.3, request_volume_window=30, reset_timeout=5, exclude=None):
        """
        :param err_threshold: float, err_rate after which circuit open
        :param request_volume_window: int, box to determine, how many request
                                are taken into consideration for calculation
                                of current error_rate
        :param reset_timeout: int, timeout in seconds
        :param exclude: List[ClassVar[Exception]] (optional), list of exception
                        classes to be ignored by the circuit breaker. It
                        essentially means that these exceptions fill be
                        transparent to circuit breaker and will not be
                        counted towards failure
        """

        def breaker_decorator(method):
            def _breaker(*args, **kwargs):
                # Used as the key to the breaker object cache
                # This will raise `KeyError` in case the function has
                # no parameter with name key. This behaviour is intentional
                cache_key = kwargs['key']

                # Used while persisting the circuit breaker state into
                # the redis
                # ignore var as we are not using redis storage as of now
                namespace = self._get_namespace(cache_key)
                # limiting Maximum size of the dict holding
                # CircuitBreaker instances based on cache_key
                size_limit = 100000
                if len(_breakers_map) >= size_limit:
                    self.logger.error("Circuit Breaker instances limit of {}"
                                      " has been exhausted".format(size_limit),
                                      exc_info=True)
                    raise exceptions.MemoryLimitExhausted(400, "Maximum number {} of Circuit Breaker"
                                                               " instances Created, dictionary size full"
                                                          .format(size_limit))

                if not _breakers_map.get(cache_key):
                    # No circuit breaker was found for the given key
                    # Lets create a new circuit breaker instance
                    _breakers_map[cache_key] = \
                        self._new_circuit_breaker(
                            err_threshold=err_threshold,
                            request_volume_window=request_volume_window,
                            reset_timeout=reset_timeout,
                            svc_name=self._service_name,
                            namespace=namespace,
                            exclude=exclude or []
                        )

                breaker = _breakers_map[cache_key]
                return breaker.call(method, *args, **kwargs)

            return _breaker

        return breaker_decorator

    def _new_circuit_breaker(self, err_threshold, request_volume_window,
                             reset_timeout, svc_name, exclude, namespace):
        return pybreaker.CircuitBreaker(
            err_threshold=err_threshold,
            request_volume_window=request_volume_window,
            reset_timeout=reset_timeout,
            svc_name=svc_name,
            exclude=exclude or [],
            listeners=[]
        )

    def _get_namespace(self, suffix=''):
        return '{}-{}'.format(self._service_name, suffix)


class _StatsdMetricReporter(pybreaker.CircuitBreakerListener):
    """
    Internal class to report circuit stats to statsd broker. Implemented
    as a listener in ``pybreaker``
    """

    def __init__(self, *args, **kwargs):
        self._metric_client = kwargs.get('metric_client')
        self._key = kwargs.get('key', '')

    def state_change(self, cb, old_state, new_state):
        if isinstance(new_state, pybreaker.CircuitClosedState):
            self._metric_client.decr(self._key, 1)
        else:
            self._metric_client.incr(self._key, 1)


class _CircuitBreakerStateLogging(pybreaker.CircuitBreakerListener):

    def __init__(self, *args, **kwargs):
        self.namespace = kwargs.get('namespace', "")
        self.logger = kwargs.get('logger')

    def state_change(self, cb, old_state, new_state):
        self.logger.error("circuit breaker state changed",
                         old_state=str(old_state._name),
                         new_state=str(new_state._name),
                         namespace=self.namespace)



from api_client import get_request_session
import urlparse

class Base:
    def __init__(self,*args,**kwargs):
        self.args = args
        self.kwargs = kwargs

    def get_user_session(self,url=None,max_retries=1):
        return get_request_session(url)

    def authenticate_user(self,user_session,user_api_user,user_api_password):
        user_session.auth = (user_api_user,user_api_password)
        return user_session
    
    def get_url(self,base_url,path=None,user_id=None):
        url = urlparse.urljoin(
            urlparse.urljoin(base_url,path,str(user_id))
        )
        return url
    




# Copyright 2013 Donald Stufft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, division, print_function
from __future__ import unicode_literals

import argparse
import os.path

from twine.commands import _find_dists
from twine.package import PackageFile
from twine import exceptions
from twine import settings
from twine import utils


def skip_upload(response, skip_existing, package):
    filename = package.basefilename
    msg_400 = (
        # Old PyPI message:
        'A file named "{}" already exists for'.format(filename),
        # Warehouse message:
        'File already exists',
        # Nexus Repository OSS message:
        'Repository does not allow updating assets',
    )
    msg_403 = 'Not enough permissions to overwrite artifact'
    # NOTE(sigmavirus24): PyPI presently returns a 400 status code with the
    # error message in the reason attribute. Other implementations return a
    # 409 or 403 status code. We only want to skip an upload if:
    # 1. The user has told us to skip existing packages (skip_existing is
    #    True) AND
    # 2. a) The response status code is 409 OR
    # 2. b) The response status code is 400 AND it has a reason that matches
    #       what we expect PyPI or Nexus OSS to return to us. OR
    # 2. c) The response status code is 403 AND the text matches what we
    #       expect Artifactory to return to us.
    return (skip_existing and (response.status_code == 409 or
            (response.status_code == 400 and
             response.reason.startswith(msg_400)) or
            (response.status_code == 403 and msg_403 in response.text)))


def upload(upload_settings, dists):
    dists = _find_dists(dists)

    # Determine if the user has passed in pre-signed distributions
    signatures = {os.path.basename(d): d for d in dists if d.endswith(".asc")}
    uploads = [i for i in dists if not i.endswith(".asc")]
    upload_settings.check_repository_url()
    repository_url = upload_settings.repository_config['repository']

    print("Uploading distributions to {}".format(repository_url))

    repository = upload_settings.create_repository()
    uploaded_packages = []

    for filename in uploads:
        package = PackageFile.from_filename(filename, upload_settings.comment)
        skip_message = (
            "  Skipping {} because it appears to already exist".format(
                package.basefilename)
        )

        # Note: The skip_existing check *needs* to be first, because otherwise
        #       we're going to generate extra HTTP requests against a hardcoded
        #       URL for no reason.
        if (upload_settings.skip_existing and
                repository.package_is_uploaded(package)):
            print(skip_message)
            continue

        signed_name = package.signed_basefilename
        if signed_name in signatures:
            package.add_gpg_signature(signatures[signed_name], signed_name)
        elif upload_settings.sign:
            package.sign(upload_settings.sign_with, upload_settings.identity)

        resp = repository.upload(package)

        # Bug 92. If we get a redirect we should abort because something seems
        # funky. The behaviour is not well defined and redirects being issued
        # by PyPI should never happen in reality. This should catch malicious
        # redirects as well.
        if resp.is_redirect:
            raise exceptions.RedirectDetected.from_args(
                repository_url,
                resp.headers["location"],
            )

        if skip_upload(resp, upload_settings.skip_existing, package):
            print(skip_message)
            continue

        utils.check_status_code(resp, upload_settings.verbose)

        uploaded_packages.append(package)

    release_urls = repository.release_urls(uploaded_packages)
    if release_urls:
        print('\nView at:')
        for url in release_urls:
            print(url)

    # Bug 28. Try to silence a ResourceWarning by clearing the connection
    # pool.
    repository.close()


def main(args):
    parser = argparse.ArgumentParser(prog="twine upload")
    settings.Settings.register_argparse_arguments(parser)
    parser.add_argument(
        "dists",
        nargs="+",
        metavar="dist",
        help="The distribution files to upload to the repository "
             "(package index). Usually dist/* . May additionally contain "
             "a .asc file to include an existing signature with the "
             "file upload.",
    )

    args = parser.parse_args(args)
    upload_settings = settings.Settings.from_argparse(args)

    # Call the upload function with the arguments from the command line
    return upload(upload_settings, args.dists)



import gtransport
from requests.packages.urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter


def get_retry_strategy(
    retries=0, status_forcelist=None, method_whitelist=None, backoff_factor=0.3
):
    if method_whitelist is None:
        method_whitelist = ["HEAD", "GET", "OPTIONS"]
    if status_forcelist is None:
        status_forcelist = [429, 500, 502, 503, 504]

    return Retry(
        total=retries,
        read=retries,
        connect=retries,
        status_forcelist=status_forcelist,
        method_whitelist=method_whitelist,
        backoff_factor=backoff_factor,
    )


def get_http_adapter(max_retries, pool_connections=10, pool_maxsize=200):
    return HTTPAdapter(
        max_retries=max_retries or get_retry_strategy(),
        pool_connections=pool_connections,
        pool_maxsize=pool_maxsize,
    )


def get_request_session(
    url=None,
    session=None,
    max_retries=0,
    adapter=None,
    pool_connections=10,
    pool_maxsize=200,
):
    """
    Wrapper function to create a session

    :param url: session for URL
    :param session: pass existing session or new session is created
    :param max_retries: int or urllib3 retry object
    :param adapter: request adapter, if not passed default http adapter is created
    :param pool_connections: int
    :param pool_maxsize: int

    :return: session object
    """
    session = session or gtransport.Session()
    max_retries = max_retries or get_retry_strategy(retries=0)

    adapter = adapter or get_http_adapter(
        max_retries=max_retries,
        pool_connections=pool_connections,
        pool_maxsize=pool_maxsize,
    )

    session.mount("http://", adapter)
    session.mount("https://", adapter)

    if url:
        session.mount(url, adapter)

    return session



# -*- coding: utf-8 -*-
"""
    pygments.lexers.grammar_notation
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Lexers for grammer notations like BNF.

    :copyright: Copyright 2006-2019 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

import re

from pygments.lexer import RegexLexer, bygroups, include, this, using, words
from pygments.token import Comment, Keyword, Literal, Name, Number, \
    Operator, Punctuation, String, Text

__all__ = ['BnfLexer', 'AbnfLexer', 'JsgfLexer']


class BnfLexer(RegexLexer):
    """
    This lexer is for grammer notations which are similar to
    original BNF.

    In order to maximize a number of targets of this lexer,
    let's decide some designs:

    * We don't distinguish `Terminal Symbol`.

    * We do assume that `NonTerminal Symbol` are always enclosed
      with arrow brackets.

    * We do assume that `NonTerminal Symbol` may include
      any printable characters except arrow brackets and ASCII 0x20.
      This assumption is for `RBNF <http://www.rfc-base.org/txt/rfc-5511.txt>`_.

    * We do assume that target notation doesn't support comment.

    * We don't distinguish any operators and punctuation except
      `::=`.

    Though these desision making might cause too minimal highlighting
    and you might be disappointed, but it is reasonable for us.

    .. versionadded:: 2.1
    """

    name = 'BNF'
    aliases = ['bnf']
    filenames = ['*.bnf']
    mimetypes = ['text/x-bnf']

    tokens = {
        'root': [
            (r'(<)([ -;=?-~]+)(>)',
             bygroups(Punctuation, Name.Class, Punctuation)),

            # an only operator
            (r'::=', Operator),

            # fallback
            (r'[^<>:]+', Text),  # for performance
            (r'.', Text),
        ],
    }


class AbnfLexer(RegexLexer):
    """
    Lexer for `IETF 7405 ABNF
    <http://www.ietf.org/rfc/rfc7405.txt>`_
    (Updates `5234 <http://www.ietf.org/rfc/rfc5234.txt>`_)
    grammars.

    .. versionadded:: 2.1
    """

    name = 'ABNF'
    aliases = ['abnf']
    filenames = ['*.abnf']
    mimetypes = ['text/x-abnf']

    _core_rules = (
        'ALPHA', 'BIT', 'CHAR', 'CR', 'CRLF', 'CTL', 'DIGIT',
        'DQUOTE', 'HEXDIG', 'HTAB', 'LF', 'LWSP', 'OCTET',
        'SP', 'VCHAR', 'WSP')

    tokens = {
        'root': [
            # comment
            (r';.*$', Comment.Single),

            # quoted
            #   double quote itself in this state, it is as '%x22'.
            (r'(%[si])?"[^"]*"', Literal),

            # binary (but i have never seen...)
            (r'%b[01]+\-[01]+\b', Literal),  # range
            (r'%b[01]+(\.[01]+)*\b', Literal),  # concat

            # decimal
            (r'%d[0-9]+\-[0-9]+\b', Literal),  # range
            (r'%d[0-9]+(\.[0-9]+)*\b', Literal),  # concat

            # hexadecimal
            (r'%x[0-9a-fA-F]+\-[0-9a-fA-F]+\b', Literal),  # range
            (r'%x[0-9a-fA-F]+(\.[0-9a-fA-F]+)*\b', Literal),  # concat

            # repetition (<a>*<b>element) including nRule
            (r'\b[0-9]+\*[0-9]+', Operator),
            (r'\b[0-9]+\*', Operator),
            (r'\b[0-9]+', Operator),
            (r'\*', Operator),

            # Strictly speaking, these are not keyword but
            # are called `Core Rule'.
            (words(_core_rules, suffix=r'\b'), Keyword),

            # nonterminals (ALPHA *(ALPHA / DIGIT / "-"))
            (r'[a-zA-Z][a-zA-Z0-9-]+\b', Name.Class),

            # operators
            (r'(=/|=|/)', Operator),

            # punctuation
            (r'[\[\]()]', Punctuation),

            # fallback
            (r'\s+', Text),
            (r'.', Text),
        ],
    }


class JsgfLexer(RegexLexer):
    """
    For `JSpeech Grammar Format <https://www.w3.org/TR/jsgf/>`_
    grammars.

    .. versionadded:: 2.2
    """
    name = 'JSGF'
    aliases = ['jsgf']
    filenames = ['*.jsgf']
    mimetypes = ['application/jsgf', 'application/x-jsgf', 'text/jsgf']

    flags = re.MULTILINE | re.UNICODE

    tokens = {
        'root': [
            include('comments'),
            include('non-comments'),
        ],
        'comments': [
            (r'/\*\*(?!/)', Comment.Multiline, 'documentation comment'),
            (r'/\*[\w\W]*?\*/', Comment.Multiline),
            (r'//.*', Comment.Single),
        ],
        'non-comments': [
            (r'\A#JSGF[^;]*', Comment.Preproc),
            (r'\s+', Text),
            (r';', Punctuation),
            (r'[=|()\[\]*+]', Operator),
            (r'/[^/]+/', Number.Float),
            (r'"', String.Double, 'string'),
            (r'\{', String.Other, 'tag'),
            (words(('import', 'public'), suffix=r'\b'), Keyword.Reserved),
            (r'grammar\b', Keyword.Reserved, 'grammar name'),
            (r'(<)(NULL|VOID)(>)',
             bygroups(Punctuation, Name.Builtin, Punctuation)),
            (r'<', Punctuation, 'rulename'),
            (r'\w+|[^\s;=|()\[\]*+/"{<\w]+', Text),
        ],
        'string': [
            (r'"', String.Double, '#pop'),
            (r'\\.', String.Escape),
            (r'[^\\"]+', String.Double),
        ],
        'tag': [
            (r'\}', String.Other, '#pop'),
            (r'\\.', String.Escape),
            (r'[^\\}]+', String.Other),
        ],
        'grammar name': [
            (r';', Punctuation, '#pop'),
            (r'\s+', Text),
            (r'\.', Punctuation),
            (r'[^;\s.]+', Name.Namespace),
        ],
        'rulename': [
            (r'>', Punctuation, '#pop'),
            (r'\*', Punctuation),
            (r'\s+', Text),
            (r'([^.>]+)(\s*)(\.)', bygroups(Name.Namespace, Text, Punctuation)),
            (r'[^.>]+', Name.Constant),
        ],
        'documentation comment': [
            (r'\*/', Comment.Multiline, '#pop'),
            (r'(^\s*\*?\s*)(@(?:example|see)\s+)'
             r'([\w\W]*?(?=(?:^\s*\*?\s*@|\*/)))',
             bygroups(Comment.Multiline, Comment.Special,
                      using(this, state='example'))),
            (r'(^\s*\*?\s*)(@\S*)',
             bygroups(Comment.Multiline, Comment.Special)),
            (r'[^*\n@]+|\w|\W', Comment.Multiline),
        ],
        'example': [
            (r'\n\s*\*', Comment.Multiline),
            include('non-comments'),
            (r'.', Comment.Multiline),
        ],
    }



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Generic utilities.
"""

from __future__ import absolute_import, division, print_function

import errno


def until_not_interrupted(f, *args, **kw):
    """
    Retry until *f* succeeds or an exception that isn't caused by EINTR occurs.

    :param callable f: A callable like a function.
    :param *args: Positional arguments for *f*.
    :param **kw: Keyword arguments for *f*.
    """
    while True:
        try:
            return f(*args, **kw)
        except (IOError, OSError) as e:
            if e.args[0] == errno.EINTR:
                continue
            raise



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Primitives to keep context global but thread (and greenlet) local.

See `thread-local`.
"""

from __future__ import absolute_import, division, print_function

import contextlib
import threading
import uuid

from structlog._config import BoundLoggerLazyProxy


try:
    from greenlet import getcurrent
except ImportError:
    from threading import local as ThreadLocal
else:
    from weakref import WeakKeyDictionary

    class ThreadLocal(object):
        """
        threading.local() replacement for greenlets.
        """

        def __init__(self):
            self.__dict__["_weakdict"] = WeakKeyDictionary()

        def __getattr__(self, name):
            key = getcurrent()
            try:
                return self._weakdict[key][name]
            except KeyError:
                raise AttributeError(name)

        def __setattr__(self, name, val):
            key = getcurrent()
            self._weakdict.setdefault(key, {})[name] = val

        def __delattr__(self, name):
            key = getcurrent()
            try:
                del self._weakdict[key][name]
            except KeyError:
                raise AttributeError(name)


def wrap_dict(dict_class):
    """
    Wrap a dict-like class and return the resulting class.

    The wrapped class and used to keep global in the current thread.

    :param type dict_class: Class used for keeping context.

    :rtype: `type`
    """
    Wrapped = type(
        "WrappedDict-" + str(uuid.uuid4()), (_ThreadLocalDictWrapper,), {}
    )
    Wrapped._tl = ThreadLocal()
    Wrapped._dict_class = dict_class
    return Wrapped


def as_immutable(logger):
    """
    Extract the context from a thread local logger into an immutable logger.

    :param structlog.BoundLogger logger: A logger with *possibly* thread local
        state.
    :rtype: :class:`~structlog.BoundLogger` with an immutable context.
    """
    if isinstance(logger, BoundLoggerLazyProxy):
        logger = logger.bind()

    try:
        ctx = logger._context._tl.dict_.__class__(logger._context._dict)
        bl = logger.__class__(
            logger._logger, processors=logger._processors, context={}
        )
        bl._context = ctx
        return bl
    except AttributeError:
        return logger


@contextlib.contextmanager
def tmp_bind(logger, **tmp_values):
    """
    Bind *tmp_values* to *logger* & memorize current state. Rewind afterwards.
    """
    saved = as_immutable(logger)._context
    try:
        yield logger.bind(**tmp_values)
    finally:
        logger._context.clear()
        logger._context.update(saved)


class _ThreadLocalDictWrapper(object):
    """
    Wrap a dict-like class and keep the state *global* but *thread-local*.

    Attempts to re-initialize only updates the wrapped dictionary.

    Useful for short-lived threaded applications like requests in web app.

    Use :func:`wrap` to instantiate and use
    :func:`structlog._loggers.BoundLogger.new` to clear the context.
    """

    def __init__(self, *args, **kw):
        """
        We cheat.  A context dict gets never recreated.
        """
        if args and isinstance(args[0], self.__class__):
            # our state is global, no need to look at args[0] if it's of our
            # class
            self._dict.update(**kw)
        else:
            self._dict.update(*args, **kw)

    @property
    def _dict(self):
        """
        Return or create and return the current context.
        """
        try:
            return self.__class__._tl.dict_
        except AttributeError:
            self.__class__._tl.dict_ = self.__class__._dict_class()
            return self.__class__._tl.dict_

    def __repr__(self):
        return "<{0}({1!r})>".format(self.__class__.__name__, self._dict)

    def __eq__(self, other):
        # Same class == same dictionary
        return self.__class__ == other.__class__

    def __ne__(self, other):
        return not self.__eq__(other)

    # Proxy methods necessary for structlog.
    # Dunder methods don't trigger __getattr__ so we need to proxy by hand.
    def __iter__(self):
        return self._dict.__iter__()

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __delitem__(self, key):
        self._dict.__delitem__(key)

    def __len__(self):
        return self._dict.__len__()

    def __getattr__(self, name):
        method = getattr(self._dict, name)
        return method


_CONTEXT = threading.local()


def merge_threadlocal(logger, method_name, event_dict):
    """
    A processor that merges in a global (thread-local) context.

    Use this as your first processor in :func:`structlog.configure` to ensure
    thread-local context is included in all log calls.

    .. versionadded:: 19.2.0

    .. versionchanged:: 20.1.0
       This function used to be called ``merge_threalocal_context`` and that
       name is still kept around for backward compatability.
    """
    context = _get_context().copy()
    context.update(event_dict)
    return context


# Alias that shouldn't be used anymore.
merge_threadlocal_context = merge_threadlocal


def clear_threadlocal():
    """
    Clear the thread-local context.

    The typical use-case for this function is to invoke it early in
    request-handling code.

    .. versionadded:: 19.2.0
    """
    _CONTEXT.context = {}


def bind_threadlocal(**kwargs):
    """
    Put keys and values into the thread-local context.

    Use this instead of :func:`~structlog.BoundLogger.bind` when you want some
    context to be global (thread-local).

    .. versionadded:: 19.2.0
    """
    _get_context().update(kwargs)


def unbind_threadlocal(*keys):
    """
    Tries to remove bound *keys* from threadlocal logging context if present.

    .. versionadded:: 20.1.0
    """
    context = _get_context()
    for key in keys:
        context.pop(key, None)


def _get_context():
    try:
        return _CONTEXT.context
    except AttributeError:
        _CONTEXT.context = {}
        return _CONTEXT.context



"""Exposes an interface for external clients"""
import logging
import threading

import requests
# proxy to requests library methods
from requests import adapters  # noqa
from requests import auth  # noqa
from requests import exceptions  # noqa
from requests.exceptions import (
    ConnectionError,
    ConnectTimeout,
    FileModeWarning,
    HTTPError,
    ReadTimeout,
    RequestException,
    Timeout,
    TooManyRedirects,
    URLRequired,
)
from requests.status_codes import codes

#  -----
import gtransport._config as _config
import gtransport._transport_client as transport

logging.getLogger(__name__).addHandler(logging.NullHandler())

_DEFAULT_PROVIDER = requests

_LOCAL = threading.local()

# The global config
_CONFIG = None


def configure(headers=None, strict=False, **cfg):
    """
    Configures the gtransport library
    @param headers: Dict[str, Any], headers to be forwarded with each
    request originating from gtransport
    @param strict: bool, flag indicating the strictness of validations
    @param cfg: optional other configuration params
    """
    global _CONFIG
    provider = cfg.get('provider') or _DEFAULT_PROVIDER
    if not provider:
        raise RuntimeError('No transport provider found')

    _CONFIG = _config.Configuration(
        provider=provider, strict=bool(strict), headers=headers, **cfg
    )


def _is_configured():
    """
    Tests whether gtransport has been configured or not.
    """
    return _CONFIG is not None


def _get_transport_client():
    """
    Creates and returns a new transport client for the current thread.
    If the client was already created for this thread, it simply
    returns the same old client.

    @return: TransportClient
    @rtype: transport.TransportClient
    """

    # Check if the library is properly configured, if not
    # configure it right about now
    if not _is_configured():
        configure()

    if not hasattr(_LOCAL, 'transport_client'):
        _LOCAL.transport_client = transport.TransportClient(
            provider=_CONFIG.provider, extra_headers=_CONFIG.headers
        )
    return _LOCAL.transport_client


def bind_transport_client(**kwargs):
    """
    Creates a new transport client for the current thread.
    It also bounds the client with a set of default headers.
    @return: TransportClient, configured instance of transport client
    @rtype: transport.TransportClient
    """
    client = _get_transport_client()
    assert isinstance(client, transport.TransportClient)

    # Fetch any extra headers to be bounded with the client
    # and associate them
    extra_headers = kwargs.get('headers') or {}
    extra_headers.update(_CONFIG.headers)
    client.update_extra_headers(retain=False, **extra_headers)
    return client


def get(url, params=None, **kwargs):
    """Convenience method to make an HTTP GET call"""
    kwargs.setdefault('allow_redirects', True)
    return _get_transport_client().send('get', url, params=params, **kwargs)


def put(url, data=None, **kwargs):
    """Convenience method to make an HTTP PUT call"""
    return _get_transport_client().send('put', url, data=data, **kwargs)


def post(url, data=None, json=None, **kwargs):
    """Convenience method to make an HTTP POST call"""
    return _get_transport_client().send('post', url, data=data, json=json, **kwargs)


def patch(url, data=None, **kwargs):
    """Convenience method to make an HTTP PATCH call"""
    return _get_transport_client().send('patch', url, data=data, **kwargs)


def delete(url, **kwargs):
    """Convenience method to make an HTTP DELETE call"""
    return _get_transport_client().send('delete', url, **kwargs)


def request(method, url, data=None, json=None, **kwargs):
    """
    A generic function to make an HTTP call
    @param method: str, the http verb (GET, HEAD, POST, PUT, etc)
    @param url: str, the URL string
    @param data: Any, the request body
    @param json: Dict[Hashable, Any], the request body as dict,
    in case the content/type was application/json
    @param kwargs: other parameters like headers, timeouts, etc. Please
    check the documentation of requests library
    """
    return _get_transport_client().send(method, url, data=data, json=json, **kwargs)


def session():
    """
    A proxy function for replicating functionality of
     requests.session() method
    """
    # importing locally to manage circular import
    from gtransport._session import GTransportSession

    return GTransportSession()


Session = session



# -*- coding: utf-8 -*-
"""
    pygments.styles.igor
    ~~~~~~~~~~~~~~~~~~~~

    Igor Pro default style.

    :copyright: Copyright 2006-2019 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.style import Style
from pygments.token import Keyword, Name, Comment, String


class IgorStyle(Style):
    """
    Pygments version of the official colors for Igor Pro procedures.
    """
    default_style = ""

    styles = {
        Comment:                'italic #FF0000',
        Keyword:                '#0000FF',
        Name.Function:          '#C34E00',
        Name.Decorator:         '#CC00A3',
        Name.Class:             '#007575',
        String:                 '#009C00'
    }



"""
Defines a transport client. A transport client is an
abstract http client
"""
import collections
import inspect
import json
import logging
import time
from urlparse import urlparse

import gtransport._utils as _utils
from gtransport._executor import basic_executor
from gtransport.constants import ERROR_THRESHOLD, REQUEST_VOLUME_WINDOW, RESET_TIMEOUT

LOGGER = logging.getLogger(__name__)


class TransportClient(object):
    """
    Represents an abstract bounded http client. We call it bounded
    because it will append a few configured headers with every request
    it sends. This gives convenience to pass headers like `request-id`,
    which is used to trace a request or a header like `service`, used
    to identify the service making the request.

    @todo: update this docstrings
    """

    _executor = None
    _provider = None

    def __init__(self, **kwargs):
        self._executor = kwargs.get('executor') or basic_executor
        self._provider = kwargs['provider']
        self._extra_headers = kwargs.get('extra_headers') or {}

    def update_extra_headers(self, retain=True, **headers):
        """
        Updates any extra headers to be appended to the request

        @param retain: bool, optional flag to determine if any
        old headers present are to be preserved. Defaults to True
        """
        headers = headers or {}
        if not bool(retain):
            self._extra_headers = {}
        self._extra_headers.update(**headers)

    def _get_service_name(self):
        svc_name = self._extra_headers.get('Service') or 'default_svc'
        return svc_name

    @staticmethod
    def _generate_key(url):
        """
        Generates a unique key for failsafe instance
        @param url: str, url includes method, scheme and netloc
        generated by urlparse lib
        url = 'https.get.xyz.example.com'
        Lets say we are in this file,
        promotions/services/espina.py
        and if we make a call like
            def recommend_products(user_id, **kwargs):
                gtransport.get(url)
        then,
        func_name = recommend_products
        module_name = services.espina
        @return: str, unique key formed
        key = https.get.xyz.example.com.services.espina.recommend_products
        """
        frame = inspect.currentframe()
        request_source = frame.f_back.f_back.f_back
        code_object = request_source.f_code
        module_name = "module_name"
        try:
            module_name = inspect.getmodule(request_source).__name__
        except AttributeError:
            LOGGER.warning("FailSafe: Using default module name - %s", module_name)

        func_name = code_object.co_name
        LOGGER.info("func_name:{} and module_name:{}".format(func_name, module_name))
        separator = '::'
        key = separator.join((url, module_name, func_name))
        return key

    @staticmethod
    def _parse_url(method, url):
        """
        Make a URL, which will contribute for forming unique key
        unique key will be used to make failsafe instance
        @param method: str, the HTTP verb
        @param url: str, the URL
            method => get
            url => 'https://xyz.example.com/report/archive/rt'
        Sample:
            for above args
            scheme = 'https'
            netloc = 'xyz.example.com'
        @return: str, URL formed with provided params
            url = 'https.get.xyz.example.com'
        """
        parsed_url = urlparse(url)
        separator = '::'
        url = separator.join((parsed_url.scheme, method, parsed_url.netloc))
        return url

    def send(self, method, url, provider=None, **kwargs):
        """
        Makes an HTTP request to the external service
        @param method: str, the HTTP verb
        @param url: str, the URL
        @param provider: Any, the low level http client, any
        object with a method: `def request(method, url, **kwargs)`,
        will do.
        @param kwargs:
        @return: same as the response from the `provider`
        """
        # Put any additional headers to be forwarded
        headers = kwargs.get('headers') or {}
        self._decorate_headers(headers)
        kwargs['headers'] = headers
        provider = provider or self._provider
        svc_name = self._get_service_name()
        parsed_url = self._parse_url(method, url)
        unique_key = kwargs.pop('key', self._generate_key(parsed_url))
        # popping out these keys is must as we cant pass these to
        # request library
        err_threshold = kwargs.pop('err_threshold', ERROR_THRESHOLD)
        request_volume_window = kwargs.pop(
            'request_volume_window', REQUEST_VOLUME_WINDOW
        )
        reset_timeout = kwargs.pop('reset_timeout', RESET_TIMEOUT)
        LOGGER.info('FailSafe key is {} for service {}'.format(unique_key, svc_name))
        return self._executor(
            func=lambda: self._request(provider, method, url, **kwargs),
            key=unique_key,
            svc_name=svc_name,
            err_threshold=err_threshold,
            request_volume_window=request_volume_window,
            reset_timeout=reset_timeout,
        )

    def _decorate_headers(self, headers):
        """Appends any additional headers to the request"""
        for name, value in self._extra_headers.items():
            headers.setdefault(name, str(value))

    @staticmethod
    def _request(provider, method, url, **kwargs):
        """
        Invokes the `request(...)` method of the given
        provider and returns the response. It also logs the
        request and response objects based on the logging level.
        """
        # dict holding contextual information to be logged
        debug = collections.OrderedDict(method=method, url=url, headers=kwargs.get('headers'))
        # Holds the result of the http call
        result = None
        LOGGER.info(
            'Upstream Request: %s'
            % json.dumps(debug, default=lambda x: repr(x))
        )
        start_time = time.time()
        try:
            result = provider.request(method, url, **kwargs)
            return result
        except Exception as ex:
            debug['exception_type'] = type(ex).__name__
            raise ex
        finally:
            end_time = time.time()
            debug.update(
                response_status=result.status_code if result else None,
                request_elapsed_time=result.elapsed.total_seconds() if result else None,
                time_taken=str(end_time - start_time)[:8],
            )

            # If debug log is enabled add request and response body
            if LOGGER.isEnabledFor(logging.DEBUG):
                debug.update(request_body=kwargs.get('data') or kwargs.get('json'))
                debug.update(response_body=result.text if result else None)
                debug.update(response_headers=result.headers if result else None)

            LOGGER.info(
                'Upstream Response: %s'
                % json.dumps(debug, default=lambda x: repr(x))
            )


class TransportClientSupplier(object):
    """A factory for the TransportClient"""

    def __init__(self, **kwargs):
        self._kw = kwargs or {}

    def __call__(self, *args, **kwargs):
        """
        Returns a new TransportClient bounded with the given context
        """
        return TransportClient(**self._kw)



# Copyright 2013 Tres Seaver
# Copyright 2015 Ian Cordasco
# This code was originally licensed under the Python Software Foudation
# License by Tres Seaver. In order to facilitate finding the metadata of
# installed packages, part of the most current implementation of the
# pkginfo.Installed class is reproduced here with bug fixes from
# https://bugs.launchpad.net/pkginfo/+bug/1437570.
import glob
import os
import sys
import warnings

import pkginfo


class Installed(pkginfo.Installed):

    def read(self):
        opj = os.path.join
        if self.package is not None:
            package = self.package.__package__
            if package is None:
                package = self.package.__name__
            egg_pattern = '%s*.egg-info' % package
            dist_pattern = '%s*.dist-info' % package
            file = getattr(self.package, '__file__', None)
            if file is not None:
                candidates = []

                def _add_candidate(where):
                    candidates.extend(glob.glob(where))

                for entry in sys.path:
                    if file.startswith(entry):
                        _add_candidate(opj(entry, 'METADATA'))  # egg?
                        _add_candidate(opj(entry, 'EGG-INFO'))  # egg?
                        # dist-installed?
                        _add_candidate(opj(entry, egg_pattern))
                        _add_candidate(opj(entry, dist_pattern))
                dir, name = os.path.split(self.package.__file__)
                _add_candidate(opj(dir, egg_pattern))
                _add_candidate(opj(dir, dist_pattern))
                _add_candidate(opj(dir, '..', egg_pattern))
                _add_candidate(opj(dir, '..', dist_pattern))

                for candidate in candidates:
                    if os.path.isdir(candidate):
                        path = opj(candidate, 'PKG-INFO')
                        if not os.path.exists(path):
                            path = opj(candidate, 'METADATA')
                    else:
                        path = candidate
                    if os.path.exists(path):
                        with open(path) as f:
                            return f.read()

        warnings.warn('No PKG-INFO or METADATA found for package: %s' %
                      self.package_name)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Structured logging for Python.
"""

from __future__ import absolute_import, division, print_function

from structlog import dev, processors, stdlib, testing, threadlocal
from structlog._base import BoundLoggerBase
from structlog._config import (
    configure,
    configure_once,
    get_config,
    get_logger,
    getLogger,
    is_configured,
    reset_defaults,
    wrap_logger,
)
from structlog._generic import BoundLogger
from structlog._loggers import PrintLogger, PrintLoggerFactory
from structlog.exceptions import DropEvent
from structlog.testing import ReturnLogger, ReturnLoggerFactory


try:
    from structlog import twisted
except ImportError:  # pragma: nocover
    twisted = None


__version__ = "20.1.0"

__title__ = "structlog"
__description__ = "Structured Logging for Python"
__uri__ = "https://www.structlog.org/"

__author__ = "Hynek Schlawack"
__email__ = "hs@ox.cx"

__license__ = "MIT or Apache License, Version 2.0"
__copyright__ = "Copyright (c) 2013 " + __author__


__all__ = [
    "BoundLogger",
    "BoundLoggerBase",
    "DropEvent",
    "PrintLogger",
    "PrintLoggerFactory",
    "ReturnLogger",
    "ReturnLoggerFactory",
    "configure",
    "configure_once",
    "dev",
    "getLogger",
    "get_config",
    "get_logger",
    "is_configured",
    "processors",
    "reset_defaults",
    "stdlib",
    "testing",
    "threadlocal",
    "twisted",
    "wrap_logger",
]



import logging
import sys


class JSONAwareLogFormatter(logging.Formatter):
    def format(self, record):
        record.message = record.getMessage()
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)
        return self._fmt % record.__dict__


class LoggerFactory(object):
    def __init__(self, level, name=None, formatter=None, **kwargs):
        self._name = name or 'glogger'
        self._fmt = formatter or JSONAwareLogFormatter(kwargs.get('fmt'))
        self._handlers = kwargs.get('handlers') or []
        self._level = level or logging.INFO
        self._stream = kwargs.get('stream') or sys.stdout
        self._logger = None

    def __call__(self, *args, **kwargs):
        if not self._logger:
            self._init_logger()
        return self._logger

    def _init_logger(self):
        sh = logging.StreamHandler(stream=self._stream)
        sh.setFormatter(self._fmt)

        logger = logging.getLogger(self._name)
        logger.setLevel(self._level)
        logger.propagate = False
        logger.addHandler(sh)
        for h in self._handlers:
            logger.addHandler(h)
        self._logger = logger



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Global state department.  Don't reload this module or everything breaks.
"""

from __future__ import absolute_import, division, print_function

import platform
import sys
import warnings

from collections import OrderedDict

from ._generic import BoundLogger
from ._loggers import PrintLoggerFactory
from .dev import ConsoleRenderer, _has_colorama, set_exc_info
from .processors import StackInfoRenderer, TimeStamper, format_exc_info


"""
.. note::

   Any changes to these defaults must be reflected in `getting-started`.
"""
_BUILTIN_DEFAULT_PROCESSORS = [
    StackInfoRenderer(),
    set_exc_info,
    format_exc_info,
    TimeStamper(fmt="%Y-%m-%d %H:%M.%S", utc=False),
    ConsoleRenderer(colors=_has_colorama),
]
if (
    sys.version_info[:2] >= (3, 6)
    or platform.python_implementation() == "PyPy"
):
    # Python 3.6+ and PyPy have ordered dicts.
    _BUILTIN_DEFAULT_CONTEXT_CLASS = dict
else:
    _BUILTIN_DEFAULT_CONTEXT_CLASS = OrderedDict
_BUILTIN_DEFAULT_WRAPPER_CLASS = BoundLogger
_BUILTIN_DEFAULT_LOGGER_FACTORY = PrintLoggerFactory()
_BUILTIN_CACHE_LOGGER_ON_FIRST_USE = False


class _Configuration(object):
    """
    Global defaults.
    """

    is_configured = False
    default_processors = _BUILTIN_DEFAULT_PROCESSORS[:]
    default_context_class = _BUILTIN_DEFAULT_CONTEXT_CLASS
    default_wrapper_class = _BUILTIN_DEFAULT_WRAPPER_CLASS
    logger_factory = _BUILTIN_DEFAULT_LOGGER_FACTORY
    cache_logger_on_first_use = _BUILTIN_CACHE_LOGGER_ON_FIRST_USE


_CONFIG = _Configuration()
"""
Global defaults used when arguments to `wrap_logger` are omitted.
"""


def is_configured():
    """
    Return whether ``structlog`` has been configured.

    If `False`, ``structlog`` is running with builtin defaults.

    :rtype: bool

    .. versionadded: 18.1
    """
    return _CONFIG.is_configured


def get_config():
    """
    Get a dictionary with the current configuration.

    .. note::

       Changes to the returned dictionary do *not* affect ``structlog``.

    :rtype: dict

    .. versionadded: 18.1
    """
    return {
        "processors": _CONFIG.default_processors,
        "context_class": _CONFIG.default_context_class,
        "wrapper_class": _CONFIG.default_wrapper_class,
        "logger_factory": _CONFIG.logger_factory,
        "cache_logger_on_first_use": _CONFIG.cache_logger_on_first_use,
    }


def get_logger(*args, **initial_values):
    """
    Convenience function that returns a logger according to configuration.

    >>> from structlog import get_logger
    >>> log = get_logger(y=23)
    >>> log.msg("hello", x=42)
    y=23 x=42 event='hello'

    :param args: *Optional* positional arguments that are passed unmodified to
        the logger factory.  Therefore it depends on the factory what they
        mean.
    :param initial_values: Values that are used to pre-populate your contexts.

    :returns: A proxy that creates a correctly configured bound logger when
        necessary.

    See `configuration` for details.

    If you prefer CamelCase, there's an alias for your reading pleasure:
    `structlog.getLogger`.

    .. versionadded:: 0.4.0
        *args*
    """
    return wrap_logger(None, logger_factory_args=args, **initial_values)


getLogger = get_logger
"""
CamelCase alias for `structlog.get_logger`.

This function is supposed to be in every source file -- we don't want it to
stick out like a sore thumb in frameworks like Twisted or Zope.
"""


def wrap_logger(
    logger,
    processors=None,
    wrapper_class=None,
    context_class=None,
    cache_logger_on_first_use=None,
    logger_factory_args=None,
    **initial_values
):
    """
    Create a new bound logger for an arbitrary *logger*.

    Default values for *processors*, *wrapper_class*, and *context_class* can
    be set using `configure`.

    If you set an attribute here, `configure` calls have *no* effect for
    the *respective* attribute.

    In other words: selective overwriting of the defaults while keeping some
    *is* possible.

    :param initial_values: Values that are used to pre-populate your contexts.
    :param tuple logger_factory_args: Values that are passed unmodified as
        ``*logger_factory_args`` to the logger factory if not `None`.

    :returns: A proxy that creates a correctly configured bound logger when
        necessary.

    See `configure` for the meaning of the rest of the arguments.

    .. versionadded:: 0.4.0
        *logger_factory_args*
    """
    return BoundLoggerLazyProxy(
        logger,
        wrapper_class=wrapper_class,
        processors=processors,
        context_class=context_class,
        cache_logger_on_first_use=cache_logger_on_first_use,
        initial_values=initial_values,
        logger_factory_args=logger_factory_args,
    )


def configure(
    processors=None,
    wrapper_class=None,
    context_class=None,
    logger_factory=None,
    cache_logger_on_first_use=None,
):
    """
    Configures the **global** defaults.

    They are used if `wrap_logger` or `get_logger` are called without
    arguments.

    Can be called several times, keeping an argument at `None` leaves it
    unchanged from the current setting.

    After calling for the first time, `is_configured` starts returning
    `True`.

    Use `reset_defaults` to undo your changes.

    :param list processors: List of processors.
    :param type wrapper_class: Class to use for wrapping loggers instead of
        `structlog.BoundLogger`.  See `standard-library`, :doc:`twisted`, and
        `custom-wrappers`.
    :param type context_class: Class to be used for internal context keeping.
    :param callable logger_factory: Factory to be called to create a new
        logger that shall be wrapped.
    :param bool cache_logger_on_first_use: `wrap_logger` doesn't return an
        actual wrapped logger but a proxy that assembles one when it's first
        used.  If this option is set to `True`, this assembled logger is
        cached.  See `performance`.

    .. versionadded:: 0.3.0
        *cache_logger_on_first_use*
    """
    _CONFIG.is_configured = True
    if processors is not None:
        _CONFIG.default_processors = processors
    if wrapper_class is not None:
        _CONFIG.default_wrapper_class = wrapper_class
    if context_class is not None:
        _CONFIG.default_context_class = context_class
    if logger_factory is not None:
        _CONFIG.logger_factory = logger_factory
    if cache_logger_on_first_use is not None:
        _CONFIG.cache_logger_on_first_use = cache_logger_on_first_use


def configure_once(*args, **kw):
    """
    Configures iff structlog isn't configured yet.

    It does *not* matter whether is was configured using `configure` or
    `configure_once` before.

    Raises a `RuntimeWarning` if repeated configuration is attempted.
    """
    if not _CONFIG.is_configured:
        configure(*args, **kw)
    else:
        warnings.warn("Repeated configuration attempted.", RuntimeWarning)


def reset_defaults():
    """
    Resets global default values to builtin defaults.

    `is_configured` starts returning `False` afterwards.
    """
    _CONFIG.is_configured = False
    _CONFIG.default_processors = _BUILTIN_DEFAULT_PROCESSORS[:]
    _CONFIG.default_wrapper_class = _BUILTIN_DEFAULT_WRAPPER_CLASS
    _CONFIG.default_context_class = _BUILTIN_DEFAULT_CONTEXT_CLASS
    _CONFIG.logger_factory = _BUILTIN_DEFAULT_LOGGER_FACTORY
    _CONFIG.cache_logger_on_first_use = _BUILTIN_CACHE_LOGGER_ON_FIRST_USE


class BoundLoggerLazyProxy(object):
    """
    Instantiates a ``BoundLogger`` on first usage.

    Takes both configuration and instantiation parameters into account.

    The only points where a BoundLogger changes state are ``bind()``,
    ``unbind()``, and ``new()`` and that return the actual ``BoundLogger``.

    If and only if configuration says so, that actual ``BoundLogger`` is
    cached on first usage.

    .. versionchanged:: 0.4.0
        Added support for *logger_factory_args*.
    """

    def __init__(
        self,
        logger,
        wrapper_class=None,
        processors=None,
        context_class=None,
        cache_logger_on_first_use=None,
        initial_values=None,
        logger_factory_args=None,
    ):
        self._logger = logger
        self._wrapper_class = wrapper_class
        self._processors = processors
        self._context_class = context_class
        self._cache_logger_on_first_use = cache_logger_on_first_use
        self._initial_values = initial_values or {}
        self._logger_factory_args = logger_factory_args or ()

    def __repr__(self):
        return (
            "<BoundLoggerLazyProxy(logger={0._logger!r}, wrapper_class="
            "{0._wrapper_class!r}, processors={0._processors!r}, "
            "context_class={0._context_class!r}, "
            "initial_values={0._initial_values!r}, "
            "logger_factory_args={0._logger_factory_args!r})>".format(self)
        )

    def bind(self, **new_values):
        """
        Assemble a new BoundLogger from arguments and configuration.
        """
        if self._context_class:
            ctx = self._context_class(self._initial_values)
        else:
            ctx = _CONFIG.default_context_class(self._initial_values)
        cls = self._wrapper_class or _CONFIG.default_wrapper_class
        _logger = self._logger
        if not _logger:
            _logger = _CONFIG.logger_factory(*self._logger_factory_args)

        if self._processors is None:
            procs = _CONFIG.default_processors
        else:
            procs = self._processors
        logger = cls(_logger, processors=procs, context=ctx)

        def finalized_bind(**new_values):
            """
            Use cached assembled logger to bind potentially new values.
            """
            if new_values:
                return logger.bind(**new_values)
            else:
                return logger

        if self._cache_logger_on_first_use is True or (
            self._cache_logger_on_first_use is None
            and _CONFIG.cache_logger_on_first_use is True
        ):
            self.bind = finalized_bind
        return finalized_bind(**new_values)

    def unbind(self, *keys):
        """
        Same as bind, except unbind *keys* first.

        In our case that could be only initial values.
        """
        return self.bind().unbind(*keys)

    def new(self, **new_values):
        """
        Clear context, then bind.
        """
        if self._context_class:
            self._context_class().clear()
        else:
            _CONFIG.default_context_class().clear()
        bl = self.bind(**new_values)
        return bl

    def __getattr__(self, name):
        """
        If a logging method if called on a lazy proxy, we have to create an
        ephemeral BoundLogger first.
        """
        if name == "__isabstractmethod__":
            raise AttributeError
        bl = self.bind()
        return getattr(bl, name)

    def __getstate__(self):
        """
        Our __getattr__ magic makes this necessary.
        """
        return self.__dict__

    def __setstate__(self, state):
        """
        Our __getattr__ magic makes this necessary.
        """
        for k, v in state.items():
            setattr(self, k, v)



import json
import logging
import sys
from functools import partial

import structlog
import util
from pygments import highlight, lexers, formatters

import lib


class SentryAwareJSONRenderer(object):
    """
    rtype: `tuple` of ``(*args, **kw)``

    This should be last processor as its call method returning tuple
    If calls json dumps on the event dict, if any of the value is not json
    serializable, then "repr" on the object is called just like JsonRenderer Processor

    JsonRenderer processor as last processor was breaking the breaking visual sentry
    stack presentation, hence new renderer is created with rtype tuple
    """
    _default_serializer = partial(
        json.dumps,
        sort_keys=True,
        default=lambda x: repr(x)
    )

    def __init__(self, serializer=None, color=False):
        self._color = color
        self._dumps = serializer or self._default_serializer

    def __call__(self, _, __, event_dict):
        data = {}

        # exc_info is required by sentry,
        # but "format_exc_info" processor removes exc_info,
        # hence setting exc_info again
        if 'exception' in event_dict:
            data['exc_info'] = True

        # as structlog will call log.msg, setting 'msg' in dict
        data['msg'] = self._dumps(event_dict)

        if self._color and util.settings.is_log_coloring_enabled():
            data['msg'] = highlight(
                data['msg'], lexers.JsonLexer(), formatters.TerminalFormatter()
            )
        return (), data


def _remove_fields(logger, method_name, event_dict):
    """
    Remove fields from log based on key 'log_request_info'
    By default log_request_info is disabled

    So, if client uses -
    logger.info('test')   # request_info is skipped.
    logger.info('test', log_request_info=True)  # request_info is rendered
    logger.error('test') # request_info is skipped
    logger.error('test', log_request_info=True)

    This will add a key named, "skip_request_info" in log if request_info is skipped
    """
    request_fields_to_remove = ['body', 'query_params', 'headers', 'host']

    def _is_enabled_request_info_log():
        event_log_request_info = event_dict.pop('log_request_info', False)
        return event_log_request_info

    if not _is_enabled_request_info_log():
        for field in request_fields_to_remove:
            event_dict.pop(field, None)

        event_dict["skip_request_info"] = True

    return event_dict


def _add_exc_info(logger, method_name, event_dict):
    """
    Add exc_info by default for error level logs
    """
    if method_name == 'error' and event_dict.get('exc_info') is None:
        event_dict['exc_info'] = True

    return event_dict


def _add_caller_runtime_info(logger, method_name, event_dict):
    '''
    Add caller's file name, line no, function name
    '''
    if not event_dict.get('file'):
        try:
            caller_file_name, caller_line_no, caller_func_name = util.find_caller()
            event_dict['file'] = caller_file_name
            event_dict['line'] = caller_line_no
            event_dict['func'] = caller_func_name
        except:
            pass

    return event_dict


class BaseStructlogConfiguration(object):
    def __init__(self, *extra_processors):
        self.processors = [
            structlog.stdlib.filter_by_level,
            structlog.stdlib.add_logger_name,
            structlog.stdlib.add_log_level,
            structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.TimeStamper(fmt="iso"),
            structlog.processors.StackInfoRenderer(),
            _add_caller_runtime_info,
            _add_exc_info,
            structlog.processors.format_exc_info,
            structlog.processors.UnicodeDecoder(),
            _remove_fields,
        ]

        self.context_class = dict
        self.logger_factory = structlog.stdlib.LoggerFactory()
        self.wrapper_class = structlog.stdlib.BoundLogger
        self.cache_logger_on_first_use = True
        for processor in extra_processors or []:
            self.processors.append(processor)

    def __call__(self, *args, **kwargs):
        self.logger_factory = lib.LoggerFactory(
            level=kwargs.get('level', logging.INFO),
            name=kwargs.get('name'),
            stream=kwargs.get('stream', sys.stdout),
            fmt="%(message)s"
        )
        structlog.configure(
            processors=self.processors,
            context_class=self.context_class,
            logger_factory=self.logger_factory,
            wrapper_class=self.wrapper_class,
            cache_logger_on_first_use=self.cache_logger_on_first_use,
        )


class StructlogDevelopmentConfiguration(BaseStructlogConfiguration):
    def __init__(self):
        super(StructlogDevelopmentConfiguration, self).__init__(
            SentryAwareJSONRenderer(color=True)
        )


class StructlogProductionConfiguration(BaseStructlogConfiguration):
    def __init__(self):
        super(StructlogProductionConfiguration, self).__init__(
            SentryAwareJSONRenderer()
        )


production_cfg = StructlogProductionConfiguration()

development_cfg = StructlogDevelopmentConfiguration()



"""Contains definition of various configuration related classes"""
from copy import copy

from gtransport import _utils
from gtransport._exceptions import StrictModeError


class Configuration(object):
    """A dict like configuration class"""

    __STRICT_HEADERS__ = {'service'}

    _strict = False

    _get_headers = None

    def __init__(self, provider, **cfg):
        """
        Supported optional parameters:

            1. headers: Dict[str, Any], map of headers to be forwarded

            2. strict: bool, A flag to ensure all validation failures
            leads to exceptions being thrown, defaults to False

        @param cfg: optional configurations as key-value pairs.
        """

        self._strict = bool(cfg.get('strict'))
        self._headers = cfg.get('headers') or {}
        self._normalize_headers()
        if self._strict:
            self._validate_headers()
        self._cfg = {'provider': provider}
        self._cfg.update(cfg or {})

    def _get_cfg(self, name, default=None):
        """
        Returns the value of config identified by the given name
        if it exists
        @param name: name of the configuration
        @param default: fallback value to return in case the config
        is not present
        @return: value of the config, or the default value if not
        present
        """
        return self._cfg.get(name, default)

    def __getitem__(self, key):
        """
        Convenience method to access a configuration value.
        This is what gives instances dict-like interface::
            >>> cfg = Configuration(id=7)
            >>> print cfg['id'] # prints 7
        """
        return self._get_cfg(key, None)

    def as_dict(self):
        """
        Returns a copy of the configuration as a Python dict
        @return Dict[Hashable, Any], dict of configurations
        """
        return copy(self._cfg)

    @property
    def headers(self):
        """
        Returns the dict of headers to be forwarded with
        the request

        @return Dict[Hashable, str], dict of headers
        """
        return self._headers

    @property
    def strict(self):
        """Convenience method to get silent flag value"""
        return self._strict

    def _normalize_headers(self):
        """
        Normalizes all the given headers
        """
        self._headers = dict([
            (_utils.normalize_header_name(k), v)
            for k, v in self._headers.items()
        ])

    def _validate_headers(self):
        """
        Validates the parameter `headers` to check if all the required
        headers in strict mode are present.
        @return: None
        @raise StrictModeError if any header is missing
        """
        not_found = []
        for name in self.__STRICT_HEADERS__:
            name = _utils.normalize_header_name(name)
            if name not in self._headers:
                not_found.append(name)

        if not_found:
            raise StrictModeError(
                'Required headers [{}] not found'.format(','.join(not_found))
            )

    @property
    def provider(self):
        """
        Convenience method to get the transport provider
        """
        return self['provider']



from __future__ import absolute_import

import re
from collections import namedtuple

from ..exceptions import LocationParseError
from ..packages import six

url_attrs = ["scheme", "auth", "host", "port", "path", "query", "fragment"]

# We only want to normalize urls with an HTTP(S) scheme.
# urllib3 infers URLs without a scheme (None) to be http.
NORMALIZABLE_SCHEMES = ("http", "https", None)

# Almost all of these patterns were derived from the
# 'rfc3986' module: https://github.com/python-hyper/rfc3986
PERCENT_RE = re.compile(r"%[a-fA-F0-9]{2}")
SCHEME_RE = re.compile(r"^(?:[a-zA-Z][a-zA-Z0-9+-]*:|/)")
URI_RE = re.compile(
    r"^(?:([a-zA-Z][a-zA-Z0-9+.-]*):)?"
    r"(?://([^\\/?#]*))?"
    r"([^?#]*)"
    r"(?:\?([^#]*))?"
    r"(?:#(.*))?$",
    re.UNICODE | re.DOTALL,
)

IPV4_PAT = r"(?:[0-9]{1,3}\.){3}[0-9]{1,3}"
HEX_PAT = "[0-9A-Fa-f]{1,4}"
LS32_PAT = "(?:{hex}:{hex}|{ipv4})".format(hex=HEX_PAT, ipv4=IPV4_PAT)
_subs = {"hex": HEX_PAT, "ls32": LS32_PAT}
_variations = [
    #                            6( h16 ":" ) ls32
    "(?:%(hex)s:){6}%(ls32)s",
    #                       "::" 5( h16 ":" ) ls32
    "::(?:%(hex)s:){5}%(ls32)s",
    # [               h16 ] "::" 4( h16 ":" ) ls32
    "(?:%(hex)s)?::(?:%(hex)s:){4}%(ls32)s",
    # [ *1( h16 ":" ) h16 ] "::" 3( h16 ":" ) ls32
    "(?:(?:%(hex)s:)?%(hex)s)?::(?:%(hex)s:){3}%(ls32)s",
    # [ *2( h16 ":" ) h16 ] "::" 2( h16 ":" ) ls32
    "(?:(?:%(hex)s:){0,2}%(hex)s)?::(?:%(hex)s:){2}%(ls32)s",
    # [ *3( h16 ":" ) h16 ] "::"    h16 ":"   ls32
    "(?:(?:%(hex)s:){0,3}%(hex)s)?::%(hex)s:%(ls32)s",
    # [ *4( h16 ":" ) h16 ] "::"              ls32
    "(?:(?:%(hex)s:){0,4}%(hex)s)?::%(ls32)s",
    # [ *5( h16 ":" ) h16 ] "::"              h16
    "(?:(?:%(hex)s:){0,5}%(hex)s)?::%(hex)s",
    # [ *6( h16 ":" ) h16 ] "::"
    "(?:(?:%(hex)s:){0,6}%(hex)s)?::",
]

UNRESERVED_PAT = r"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789._!\-~"
IPV6_PAT = "(?:" + "|".join([x % _subs for x in _variations]) + ")"
ZONE_ID_PAT = "(?:%25|%)(?:[" + UNRESERVED_PAT + "]|%[a-fA-F0-9]{2})+"
IPV6_ADDRZ_PAT = r"\[" + IPV6_PAT + r"(?:" + ZONE_ID_PAT + r")?\]"
REG_NAME_PAT = r"(?:[^\[\]%:/?#]|%[a-fA-F0-9]{2})*"
TARGET_RE = re.compile(r"^(/[^?#]*)(?:\?([^#]*))?(?:#.*)?$")

IPV4_RE = re.compile("^" + IPV4_PAT + "$")
IPV6_RE = re.compile("^" + IPV6_PAT + "$")
IPV6_ADDRZ_RE = re.compile("^" + IPV6_ADDRZ_PAT + "$")
BRACELESS_IPV6_ADDRZ_RE = re.compile("^" + IPV6_ADDRZ_PAT[2:-2] + "$")
ZONE_ID_RE = re.compile("(" + ZONE_ID_PAT + r")\]$")

SUBAUTHORITY_PAT = (u"^(?:(.*)@)?(%s|%s|%s)(?::([0-9]{0,5}))?$") % (
    REG_NAME_PAT,
    IPV4_PAT,
    IPV6_ADDRZ_PAT,
)
SUBAUTHORITY_RE = re.compile(SUBAUTHORITY_PAT, re.UNICODE | re.DOTALL)

UNRESERVED_CHARS = set(
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789._-~"
)
SUB_DELIM_CHARS = set("!$&'()*+,;=")
USERINFO_CHARS = UNRESERVED_CHARS | SUB_DELIM_CHARS | {":"}
PATH_CHARS = USERINFO_CHARS | {"@", "/"}
QUERY_CHARS = FRAGMENT_CHARS = PATH_CHARS | {"?"}


class Url(namedtuple("Url", url_attrs)):
    """
    Data structure for representing an HTTP URL. Used as a return value for
    :func:`parse_url`. Both the scheme and host are normalized as they are
    both case-insensitive according to RFC 3986.
    """

    __slots__ = ()

    def __new__(
        cls,
        scheme=None,
        auth=None,
        host=None,
        port=None,
        path=None,
        query=None,
        fragment=None,
    ):
        if path and not path.startswith("/"):
            path = "/" + path
        if scheme is not None:
            scheme = scheme.lower()
        return super(Url, cls).__new__(
            cls, scheme, auth, host, port, path, query, fragment
        )

    @property
    def hostname(self):
        """For backwards-compatibility with urlparse. We're nice like that."""
        return self.host

    @property
    def request_uri(self):
        """Absolute path including the query string."""
        uri = self.path or "/"

        if self.query is not None:
            uri += "?" + self.query

        return uri

    @property
    def netloc(self):
        """Network location including host and port"""
        if self.port:
            return "%s:%d" % (self.host, self.port)
        return self.host

    @property
    def url(self):
        """
        Convert self into a url

        This function should more or less round-trip with :func:`.parse_url`. The
        returned url may not be exactly the same as the url inputted to
        :func:`.parse_url`, but it should be equivalent by the RFC (e.g., urls
        with a blank port will have : removed).

        Example: ::

            >>> U = parse_url('http://google.com/mail/')
            >>> U.url
            'http://google.com/mail/'
            >>> Url('http', 'username:password', 'host.com', 80,
            ... '/path', 'query', 'fragment').url
            'http://username:password@host.com:80/path?query#fragment'
        """
        scheme, auth, host, port, path, query, fragment = self
        url = u""

        # We use "is not None" we want things to happen with empty strings (or 0 port)
        if scheme is not None:
            url += scheme + u"://"
        if auth is not None:
            url += auth + u"@"
        if host is not None:
            url += host
        if port is not None:
            url += u":" + str(port)
        if path is not None:
            url += path
        if query is not None:
            url += u"?" + query
        if fragment is not None:
            url += u"#" + fragment

        return url

    def __str__(self):
        return self.url


def split_first(s, delims):
    """
    .. deprecated:: 1.25

    Given a string and an iterable of delimiters, split on the first found
    delimiter. Return two split parts and the matched delimiter.

    If not found, then the first part is the full input string.

    Example::

        >>> split_first('foo/bar?baz', '?/=')
        ('foo', 'bar?baz', '/')
        >>> split_first('foo/bar?baz', '123')
        ('foo/bar?baz', '', None)

    Scales linearly with number of delims. Not ideal for large number of delims.
    """
    min_idx = None
    min_delim = None
    for d in delims:
        idx = s.find(d)
        if idx < 0:
            continue

        if min_idx is None or idx < min_idx:
            min_idx = idx
            min_delim = d

    if min_idx is None or min_idx < 0:
        return s, "", None

    return s[:min_idx], s[min_idx + 1 :], min_delim


def _encode_invalid_chars(component, allowed_chars, encoding="utf-8"):
    """Percent-encodes a URI component without reapplying
    onto an already percent-encoded component.
    """
    if component is None:
        return component

    component = six.ensure_text(component)

    # Normalize existing percent-encoded bytes.
    # Try to see if the component we're encoding is already percent-encoded
    # so we can skip all '%' characters but still encode all others.
    component, percent_encodings = PERCENT_RE.subn(
        lambda match: match.group(0).upper(), component
    )

    uri_bytes = component.encode("utf-8", "surrogatepass")
    is_percent_encoded = percent_encodings == uri_bytes.count(b"%")
    encoded_component = bytearray()

    for i in range(0, len(uri_bytes)):
        # Will return a single character bytestring on both Python 2 & 3
        byte = uri_bytes[i : i + 1]
        byte_ord = ord(byte)
        if (is_percent_encoded and byte == b"%") or (
            byte_ord < 128 and byte.decode() in allowed_chars
        ):
            encoded_component += byte
            continue
        encoded_component.extend(b"%" + (hex(byte_ord)[2:].encode().zfill(2).upper()))

    return encoded_component.decode(encoding)


def _remove_path_dot_segments(path):
    # See http://tools.ietf.org/html/rfc3986#section-5.2.4 for pseudo-code
    segments = path.split("/")  # Turn the path into a list of segments
    output = []  # Initialize the variable to use to store output

    for segment in segments:
        # '.' is the current directory, so ignore it, it is superfluous
        if segment == ".":
            continue
        # Anything other than '..', should be appended to the output
        elif segment != "..":
            output.append(segment)
        # In this case segment == '..', if we can, we should pop the last
        # element
        elif output:
            output.pop()

    # If the path starts with '/' and the output is empty or the first string
    # is non-empty
    if path.startswith("/") and (not output or output[0]):
        output.insert(0, "")

    # If the path starts with '/.' or '/..' ensure we add one more empty
    # string to add a trailing '/'
    if path.endswith(("/.", "/..")):
        output.append("")

    return "/".join(output)


def _normalize_host(host, scheme):
    if host:
        if isinstance(host, six.binary_type):
            host = six.ensure_str(host)

        if scheme in NORMALIZABLE_SCHEMES:
            is_ipv6 = IPV6_ADDRZ_RE.match(host)
            if is_ipv6:
                match = ZONE_ID_RE.search(host)
                if match:
                    start, end = match.span(1)
                    zone_id = host[start:end]

                    if zone_id.startswith("%25") and zone_id != "%25":
                        zone_id = zone_id[3:]
                    else:
                        zone_id = zone_id[1:]
                    zone_id = "%" + _encode_invalid_chars(zone_id, UNRESERVED_CHARS)
                    return host[:start].lower() + zone_id + host[end:]
                else:
                    return host.lower()
            elif not IPV4_RE.match(host):
                return six.ensure_str(
                    b".".join([_idna_encode(label) for label in host.split(".")])
                )
    return host


def _idna_encode(name):
    if name and any([ord(x) > 128 for x in name]):
        try:
            import idna
        except ImportError:
            six.raise_from(
                LocationParseError("Unable to parse URL without the 'idna' module"),
                None,
            )
        try:
            return idna.encode(name.lower(), strict=True, std3_rules=True)
        except idna.IDNAError:
            six.raise_from(
                LocationParseError(u"Name '%s' is not a valid IDNA label" % name), None
            )
    return name.lower().encode("ascii")


def _encode_target(target):
    """Percent-encodes a request target so that there are no invalid characters"""
    path, query = TARGET_RE.match(target).groups()
    target = _encode_invalid_chars(path, PATH_CHARS)
    query = _encode_invalid_chars(query, QUERY_CHARS)
    if query is not None:
        target += "?" + query
    return target


def parse_url(url):
    """
    Given a url, return a parsed :class:`.Url` namedtuple. Best-effort is
    performed to parse incomplete urls. Fields not provided will be None.
    This parser is RFC 3986 compliant.

    The parser logic and helper functions are based heavily on
    work done in the ``rfc3986`` module.

    :param str url: URL to parse into a :class:`.Url` namedtuple.

    Partly backwards-compatible with :mod:`urlparse`.

    Example::

        >>> parse_url('http://google.com/mail/')
        Url(scheme='http', host='google.com', port=None, path='/mail/', ...)
        >>> parse_url('google.com:80')
        Url(scheme=None, host='google.com', port=80, path=None, ...)
        >>> parse_url('/foo?bar')
        Url(scheme=None, host=None, port=None, path='/foo', query='bar', ...)
    """
    if not url:
        # Empty
        return Url()

    source_url = url
    if not SCHEME_RE.search(url):
        url = "//" + url

    try:
        scheme, authority, path, query, fragment = URI_RE.match(url).groups()
        normalize_uri = scheme is None or scheme.lower() in NORMALIZABLE_SCHEMES

        if scheme:
            scheme = scheme.lower()

        if authority:
            auth, host, port = SUBAUTHORITY_RE.match(authority).groups()
            if auth and normalize_uri:
                auth = _encode_invalid_chars(auth, USERINFO_CHARS)
            if port == "":
                port = None
        else:
            auth, host, port = None, None, None

        if port is not None:
            port = int(port)
            if not (0 <= port <= 65535):
                raise LocationParseError(url)

        host = _normalize_host(host, scheme)

        if normalize_uri and path:
            path = _remove_path_dot_segments(path)
            path = _encode_invalid_chars(path, PATH_CHARS)
        if normalize_uri and query:
            query = _encode_invalid_chars(query, QUERY_CHARS)
        if normalize_uri and fragment:
            fragment = _encode_invalid_chars(fragment, FRAGMENT_CHARS)

    except (ValueError, AttributeError):
        return six.raise_from(LocationParseError(source_url), None)

    # For the sake of backwards compatibility we put empty
    # string values for path if there are any defined values
    # beyond the path in the URL.
    # TODO: Remove this when we break backwards compatibility.
    if not path:
        if query is not None or fragment is not None:
            path = ""
        else:
            path = None

    # Ensure that each part of the URL is a `str` for
    # backwards compatibility.
    if isinstance(url, six.text_type):
        ensure_func = six.ensure_text
    else:
        ensure_func = six.ensure_str

    def ensure_type(x):
        return x if x is None else ensure_func(x)

    return Url(
        scheme=ensure_type(scheme),
        auth=ensure_type(auth),
        host=ensure_type(host),
        port=port,
        path=ensure_type(path),
        query=ensure_type(query),
        fragment=ensure_type(fragment),
    )


def get_host(url):
    """
    Deprecated. Use :func:`parse_url` instead.
    """
    p = parse_url(url)
    return p.scheme or "http", p.hostname, p.port



"""
Contains a set of utility functions to validate
input arguments
"""


def not_none(obj, msg):
    """
    Raises ValueError in case the given input is None
    @param x: Any, the input value
    @param msg: str, message in the raised Exception (if any)
    @return: None
    """
    if obj is None:
        raise ValueError(msg)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Helpers that make development with ``structlog`` more pleasant.
"""

from __future__ import absolute_import, division, print_function

import sys

from six import PY2, StringIO, string_types


try:
    import colorama
except ImportError:
    colorama = None


__all__ = ["ConsoleRenderer"]

_IS_WINDOWS = sys.platform == "win32"

_MISSING = "{who} requires the {package} package installed.  "
_EVENT_WIDTH = 30  # pad the event name to so many characters


def _pad(s, l):
    """
    Pads *s* to length *l*.
    """
    missing = l - len(s)
    return s + " " * (missing if missing > 0 else 0)


if colorama is not None:
    _has_colorama = True

    RESET_ALL = colorama.Style.RESET_ALL
    BRIGHT = colorama.Style.BRIGHT
    DIM = colorama.Style.DIM
    RED = colorama.Fore.RED
    BLUE = colorama.Fore.BLUE
    CYAN = colorama.Fore.CYAN
    MAGENTA = colorama.Fore.MAGENTA
    YELLOW = colorama.Fore.YELLOW
    GREEN = colorama.Fore.GREEN
    RED_BACK = colorama.Back.RED
else:
    _has_colorama = False

    RESET_ALL = (
        BRIGHT
    ) = DIM = RED = BLUE = CYAN = MAGENTA = YELLOW = GREEN = RED_BACK = ""


class _ColorfulStyles(object):
    reset = RESET_ALL
    bright = BRIGHT

    level_critical = RED
    level_exception = RED
    level_error = RED
    level_warn = YELLOW
    level_info = GREEN
    level_debug = GREEN
    level_notset = RED_BACK

    timestamp = DIM
    logger_name = BLUE
    kv_key = CYAN
    kv_value = MAGENTA


class _PlainStyles(object):
    reset = ""
    bright = ""

    level_critical = ""
    level_exception = ""
    level_error = ""
    level_warn = ""
    level_info = ""
    level_debug = ""
    level_notset = ""

    timestamp = ""
    logger_name = ""
    kv_key = ""
    kv_value = ""


class ConsoleRenderer(object):
    """
    Render ``event_dict`` nicely aligned, possibly in colors, and ordered.

    If ``event_dict`` contains an ``exception`` key (for example from
    :func:`~structlog.processors.format_exc_info`), it will be rendered *after*
    the log line.

    :param int pad_event: Pad the event to this many characters.
    :param bool colors: Use colors for a nicer output.
    :param bool force_colors: Force colors even for non-tty destinations.
        Use this option if your logs are stored in a file that is meant
        to be streamed to the console.
    :param bool repr_native_str: When `True`, `repr` is also applied
        to native strings (i.e. unicode on Python 3 and bytes on Python 2).
        Setting this to `False` is useful if you want to have human-readable
        non-ASCII output on Python 2.  The ``event`` key is *never*
        `repr` -ed.
    :param dict level_styles: When present, use these styles for colors. This
        must be a dict from level names (strings) to colorama styles. The
        default can be obtained by calling
        `ConsoleRenderer.get_default_level_styles`

    Requires the colorama_ package if *colors* is `True`.

    .. _colorama: https://pypi.org/project/colorama/

    .. versionadded:: 16.0
    .. versionadded:: 16.1 *colors*
    .. versionadded:: 17.1 *repr_native_str*
    .. versionadded:: 18.1 *force_colors*
    .. versionadded:: 18.1 *level_styles*
    .. versionchanged:: 19.2
       ``colorama`` now initializes lazily to avoid unwanted initializations as
       ``ConsoleRenderer`` is used by default.
    .. versionchanged:: 19.2 Can be pickled now.
    .. versionchanged:: 20.1 ``colorama`` does not initialize lazily on Windows
       anymore because it breaks rendering.
    """

    def __init__(
        self,
        pad_event=_EVENT_WIDTH,
        colors=_has_colorama,
        force_colors=False,
        repr_native_str=False,
        level_styles=None,
    ):
        self._force_colors = self._init_colorama = False
        if colors is True:
            if colorama is None:
                raise SystemError(
                    _MISSING.format(
                        who=self.__class__.__name__ + " with `colors=True`",
                        package="colorama",
                    )
                )

            if _IS_WINDOWS:  # pragma: no cover
                _init_colorama(self._force_colors)
            else:
                self._init_colorama = True
                if force_colors:
                    self._force_colors = True

            styles = _ColorfulStyles
        else:
            styles = _PlainStyles

        self._styles = styles
        self._pad_event = pad_event

        if level_styles is None:
            self._level_to_color = self.get_default_level_styles(colors)
        else:
            self._level_to_color = level_styles

        for key in self._level_to_color.keys():
            self._level_to_color[key] += styles.bright
        self._longest_level = len(
            max(self._level_to_color.keys(), key=lambda e: len(e))
        )

        self._repr_native_str = repr_native_str

    def _repr(self, val):
        """
        Determine representation of *val* depending on its type &
        self._repr_native_str.
        """
        if self._repr_native_str is True:
            return repr(val)

        if isinstance(val, str):
            return val
        else:
            return repr(val)

    def __call__(self, _, __, event_dict):
        # Initialize lazily to prevent import side-effects.
        if self._init_colorama:
            _init_colorama(self._force_colors)
            self._init_colorama = False
        sio = StringIO()

        ts = event_dict.pop("timestamp", None)
        if ts is not None:
            sio.write(
                # can be a number if timestamp is UNIXy
                self._styles.timestamp
                + str(ts)
                + self._styles.reset
                + " "
            )
        level = event_dict.pop("level", None)
        if level is not None:
            sio.write(
                "["
                + self._level_to_color[level]
                + _pad(level, self._longest_level)
                + self._styles.reset
                + "] "
            )

        # force event to str for compatibility with standard library
        event = event_dict.pop("event")
        if not PY2 or not isinstance(event, string_types):
            event = str(event)

        if event_dict:
            event = _pad(event, self._pad_event) + self._styles.reset + " "
        else:
            event += self._styles.reset
        sio.write(self._styles.bright + event)

        logger_name = event_dict.pop("logger", None)
        if logger_name is not None:
            sio.write(
                "["
                + self._styles.logger_name
                + self._styles.bright
                + logger_name
                + self._styles.reset
                + "] "
            )

        stack = event_dict.pop("stack", None)
        exc = event_dict.pop("exception", None)
        sio.write(
            " ".join(
                self._styles.kv_key
                + key
                + self._styles.reset
                + "="
                + self._styles.kv_value
                + self._repr(event_dict[key])
                + self._styles.reset
                for key in sorted(event_dict.keys())
            )
        )

        if stack is not None:
            sio.write("\n" + stack)
            if exc is not None:
                sio.write("\n\n" + "=" * 79 + "\n")
        if exc is not None:
            sio.write("\n" + exc)

        return sio.getvalue()

    @staticmethod
    def get_default_level_styles(colors=True):
        """
        Get the default styles for log levels

        This is intended to be used with `ConsoleRenderer`'s ``level_styles``
        parameter.  For example, if you are adding custom levels in your
        home-grown :func:`~structlog.stdlib.add_log_level` you could do::

            my_styles = ConsoleRenderer.get_default_level_styles()
            my_styles["EVERYTHING_IS_ON_FIRE"] = my_styles["critical"]
            renderer = ConsoleRenderer(level_styles=my_styles)

        :param bool colors: Whether to use colorful styles. This must match the
            *colors* parameter to `ConsoleRenderer`. Default: `True`.
        """
        if colors:
            styles = _ColorfulStyles
        else:
            styles = _PlainStyles
        return {
            "critical": styles.level_critical,
            "exception": styles.level_exception,
            "error": styles.level_error,
            "warn": styles.level_warn,
            "warning": styles.level_warn,
            "info": styles.level_info,
            "debug": styles.level_debug,
            "notset": styles.level_notset,
        }


def _init_colorama(force):
    if force:
        colorama.deinit()
        colorama.init(strip=False)
    else:
        colorama.init()


_SENTINEL = object()


def set_exc_info(_, method_name, event_dict):
    """
    Set ``event_dict["exc_info"] = True`` if *method_name* is ``"exception"``.

    Do nothing if the name is different or ``exc_info`` is already set.
    """
    if (
        method_name != "exception"
        or event_dict.get("exc_info", _SENTINEL) is not _SENTINEL
    ):
        return event_dict

    event_dict["exc_info"] = True

    return event_dict



# -*- coding: utf-8 -*-
"""A X509Adapter for use with the requests library.

This file contains an implementation of the X509Adapter that will
allow users to authenticate a request using an arbitrary
X.509 certificate without needing to convert it to a .pem file

"""

from OpenSSL.crypto import PKey, X509
from cryptography import x509
from cryptography.hazmat.primitives.serialization import (load_pem_private_key,
                                                          load_der_private_key)
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.backends import default_backend

from datetime import datetime
from requests.adapters import HTTPAdapter
import requests

from .._compat import PyOpenSSLContext
from .. import exceptions as exc

"""
importing the protocol constants from _ssl instead of ssl because only the
constants are needed and to handle issues caused by importing from ssl on
the 2.7.x line.
"""
try:
    from _ssl import PROTOCOL_TLS as PROTOCOL
except ImportError:
    from _ssl import PROTOCOL_SSLv23 as PROTOCOL


class X509Adapter(HTTPAdapter):
    r"""Adapter for use with X.509 certificates.

    Provides an interface for Requests sessions to contact HTTPS urls and
    authenticate  with an X.509 cert by implementing the Transport Adapter
    interface. This class will need to be manually instantiated and mounted
    to the session

    :param pool_connections: The number of urllib3 connection pools to
           cache.
    :param pool_maxsize: The maximum number of connections to save in the
            pool.
    :param max_retries: The maximum number of retries each connection
        should attempt. Note, this applies only to failed DNS lookups,
        socket connections and connection timeouts, never to requests where
        data has made it to the server. By default, Requests does not retry
        failed connections. If you need granular control over the
        conditions under which we retry a request, import urllib3's
        ``Retry`` class and pass that instead.
    :param pool_block: Whether the connection pool should block for
            connections.

    :param bytes cert_bytes:
        bytes object containing contents of a cryptography.x509Certificate
        object using the encoding specified by the ``encoding`` parameter.
    :param bytes pk_bytes:
        bytes object containing contents of a object that implements
        ``cryptography.hazmat.primitives.serialization.PrivateFormat``
        using the encoding specified by the ``encoding`` parameter.
    :param password:
        string or utf8 encoded bytes containing the passphrase used for the
        private key. None if unencrypted. Defaults to None.
    :param encoding:
        Enumeration detailing the encoding method used on the ``cert_bytes``
        parameter. Can be either PEM or DER. Defaults to PEM.
    :type encoding:
        :class: `cryptography.hazmat.primitives.serialization.Encoding`

    Usage::

      >>> import requests
      >>> from requests_toolbelt.adapters.x509 import X509Adapter
      >>> s = requests.Session()
      >>> a = X509Adapter(max_retries=3,
                cert_bytes=b'...', pk_bytes=b'...', encoding='...'
      >>> s.mount('https://', a)
    """

    def __init__(self, *args, **kwargs):
        self._check_version()
        cert_bytes = kwargs.pop('cert_bytes', None)
        pk_bytes = kwargs.pop('pk_bytes', None)
        password = kwargs.pop('password', None)
        encoding = kwargs.pop('encoding', Encoding.PEM)

        password_bytes = None

        if cert_bytes is None or not isinstance(cert_bytes, bytes):
            raise ValueError('Invalid cert content provided. '
                             'You must provide an X.509 cert '
                             'formatted as a byte array.')
        if pk_bytes is None or not isinstance(pk_bytes, bytes):
            raise ValueError('Invalid private key content provided. '
                             'You must provide a private key '
                             'formatted as a byte array.')

        if isinstance(password, bytes):
            password_bytes = password
        elif password:
            password_bytes = password.encode('utf8')

        self.ssl_context = create_ssl_context(cert_bytes, pk_bytes,
                                              password_bytes, encoding)

        super(X509Adapter, self).__init__(*args, **kwargs)

    def init_poolmanager(self, *args, **kwargs):
        if self.ssl_context:
            kwargs['ssl_context'] = self.ssl_context
        return super(X509Adapter, self).init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args, **kwargs):
        if self.ssl_context:
            kwargs['ssl_context'] = self.ssl_context
        return super(X509Adapter, self).proxy_manager_for(*args, **kwargs)

    def _check_version(self):
        if PyOpenSSLContext is None:
            raise exc.VersionMismatchError(
                "The X509Adapter requires at least Requests 2.12.0 to be "
                "installed. Version {0} was found instead.".format(
                    requests.__version__
                )
            )


def check_cert_dates(cert):
    """Verify that the supplied client cert is not invalid."""

    now = datetime.utcnow()
    if cert.not_valid_after < now or cert.not_valid_before > now:
        raise ValueError('Client certificate expired: Not After: '
                         '{0:%Y-%m-%d %H:%M:%SZ} '
                         'Not Before: {1:%Y-%m-%d %H:%M:%SZ}'
                         .format(cert.not_valid_after, cert.not_valid_before))


def create_ssl_context(cert_byes, pk_bytes, password=None,
                       encoding=Encoding.PEM):
    """Create an SSL Context with the supplied cert/password.

    :param cert_bytes array of bytes containing the cert encoded
           using the method supplied in the ``encoding`` parameter
    :param pk_bytes array of bytes containing the private key encoded
           using the method supplied in the ``encoding`` parameter
    :param password array of bytes containing the passphrase to be used
           with the supplied private key. None if unencrypted.
           Defaults to None.
    :param encoding ``cryptography.hazmat.primitives.serialization.Encoding``
            details the encoding method used on the ``cert_bytes``  and
            ``pk_bytes`` parameters. Can be either PEM or DER.
            Defaults to PEM.
    """
    backend = default_backend()

    cert = None
    key = None
    if encoding == Encoding.PEM:
        cert = x509.load_pem_x509_certificate(cert_byes, backend)
        key = load_pem_private_key(pk_bytes, password, backend)
    elif encoding == Encoding.DER:
        cert = x509.load_der_x509_certificate(cert_byes, backend)
        key = load_der_private_key(pk_bytes, password, backend)
    else:
        raise ValueError('Invalid encoding provided: Must be PEM or DER')

    if not (cert and key):
        raise ValueError('Cert and key could not be parsed from '
                         'provided data')
    check_cert_dates(cert)
    ssl_context = PyOpenSSLContext(PROTOCOL)
    ssl_context._ctx.use_certificate(X509.from_cryptography(cert))
    ssl_context._ctx.use_privatekey(PKey.from_cryptography_key(key))
    return ssl_context



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Primitives to deal with a concurrency supporting context, as introduced in
Python 3.7 as :mod:`contextvars`.

.. versionadded:: 20.1.0

See :doc:`contextvars`.
"""

from __future__ import absolute_import, division, print_function

import contextvars


_CONTEXT = contextvars.ContextVar("structlog_context")


def merge_contextvars(logger, method_name, event_dict):
    """
    A processor that merges in a global (context-local) context.

    Use this as your first processor in :func:`structlog.configure` to ensure
    context-local context is included in all log calls.

    .. versionadded:: 20.1.0
    """
    ctx = _get_context().copy()
    ctx.update(event_dict)
    return ctx


def clear_contextvars():
    """
    Clear the context-local context.

    The typical use-case for this function is to invoke it early in request-
    handling code.

    .. versionadded:: 20.1.0
    """
    ctx = _get_context()
    ctx.clear()


def bind_contextvars(**kwargs):
    """
    Put keys and values into the context-local context.

    Use this instead of :func:`~structlog.BoundLogger.bind` when you want some
    context to be global (context-local).

    .. versionadded:: 20.1.0
    """
    _get_context().update(kwargs)


def unbind_contextvars(*keys):
    """
    Remove *keys* from the context-local context if they are present.

    Use this instead of :func:`~structlog.BoundLogger.unbind` when you want to
    remove keys from a global (context-local) context.

    .. versionadded:: 20.1.0
    """
    ctx = _get_context()
    for key in keys:
        ctx.pop(key, None)


def _get_context():
    try:
        return _CONTEXT.get()
    except LookupError:
        _CONTEXT.set({})
        return _CONTEXT.get()



import requests

from gtransport import _get_transport_client


class GTransportSession(requests.Session):
    def __init__(self):
        super(GTransportSession, self).__init__()
        self.parent = super(GTransportSession, self)

    def get(self, url, params=None, **kwargs):
        kwargs.setdefault('allow_redirects', True)
        return _get_transport_client().send(
            'get', url, provider=self.parent, params=params, **kwargs
        )

    def put(self, url, data=None, **kwargs):
        """Convenience method to make an HTTP PUT call"""
        return _get_transport_client().send(
            'put', url, provider=self.parent, data=data, **kwargs
        )

    def post(self, url, data=None, json=None, **kwargs):
        """Convenience method to make an HTTP POST call"""
        return _get_transport_client().send(
            'post', url, provider=self.parent, data=data, json=json, **kwargs
        )

    def patch(self, url, data=None, **kwargs):
        """Convenience method to make an HTTP PATCH call"""
        return _get_transport_client().send(
            'patch', url, provider=self.parent, data=data, **kwargs
        )

    def delete(self, url, **kwargs):
        """Convenience method to make an HTTP DELETE call"""
        return _get_transport_client().send(
            'delete', url, provider=self.parent, **kwargs
        )

    def request(self, method, url, data=None, json=None, **kwargs):
        """
        A generic function to make an HTTP call
        """
        return _get_transport_client().send(
            method, url, provider=self.parent, data=data, json=json, **kwargs
        )



from __future__ import absolute_import, division, unicode_literals

import random
from collections import deque
from datetime import timedelta

from .timer import Timer


class StatsClientBase(object):
    """A Base class for various statsd clients."""

    def _send(self):
        raise NotImplementedError()

    def pipeline(self):
        raise NotImplementedError()

    def timer(self, stat, rate=1):
        return Timer(self, stat, rate)

    def timing(self, stat, delta, rate=1):
        """
        Send new timing information.

        `delta` can be either a number of milliseconds or a timedelta.
        """
        if isinstance(delta, timedelta):
            # Convert timedelta to number of milliseconds.
            delta = delta.total_seconds() * 1000.
        self._send_stat(stat, '%0.6f|ms' % delta, rate)

    def incr(self, stat, count=1, rate=1):
        """Increment a stat by `count`."""
        self._send_stat(stat, '%s|c' % count, rate)

    def decr(self, stat, count=1, rate=1):
        """Decrement a stat by `count`."""
        self.incr(stat, -count, rate)

    def gauge(self, stat, value, rate=1, delta=False):
        """Set a gauge value."""
        if value < 0 and not delta:
            if rate < 1:
                if random.random() > rate:
                    return
            with self.pipeline() as pipe:
                pipe._send_stat(stat, '0|g', 1)
                pipe._send_stat(stat, '%s|g' % value, 1)
        else:
            prefix = '+' if delta and value >= 0 else ''
            self._send_stat(stat, '%s%s|g' % (prefix, value), rate)

    def set(self, stat, value, rate=1):
        """Set a set value."""
        self._send_stat(stat, '%s|s' % value, rate)

    def _send_stat(self, stat, value, rate):
        self._after(self._prepare(stat, value, rate))

    def _prepare(self, stat, value, rate):
        if rate < 1:
            if random.random() > rate:
                return
            value = '%s|@%s' % (value, rate)

        if self._prefix:
            stat = '%s.%s' % (self._prefix, stat)

        return '%s:%s' % (stat, value)

    def _after(self, data):
        if data:
            self._send(data)


class PipelineBase(StatsClientBase):

    def __init__(self, client):
        self._client = client
        self._prefix = client._prefix
        self._stats = deque()

    def _send(self):
        raise NotImplementedError()

    def _after(self, data):
        if data is not None:
            self._stats.append(data)

    def __enter__(self):
        return self

    def __exit__(self, typ, value, tb):
        self.send()

    def send(self):
        if not self._stats:
            return
        self._send()

    def pipeline(self):
        return self.__class__(self)
